local BestInSlot = LibStub("AceAddon-3.0"):GetAddon("BestInSlot")
local AceGUI = LibStub("AceGUI-3.0")
local L = LibStub("AceLocale-3.0"):GetLocale("BestInSlot")
local eventId = 0
local eventHandlers = {}
---
--List of events that the add-on makes
-- InitializeLootTable: Fired when a loot table is initialized
--    arg1 : eventName
--    arg2 : ID of the raid Tier that's initialized
-- 
-- DebugOptionsChanged: Fired when the debug options have changed
--    arg1 :  True/false wether the options are now true or false 
-- 
-- Any AddonEvent Message received will automatically fire an event in the format
-- AddonMessage_Identifier (for example: AddonMessage_version when a version message is received)
-- 
-- Any AddonEvent Message sent will automatically fire an event in the format
-- MessageSent_Identifier (for example: MessageSent_versioncheck when a version message is received)
---



---Checks if the callback is allready present
--@param #string name The name of the callback
--@param #function callback The function to call
local function callbackExists(name, callback)
  if not eventHandlers[name] then return false end
  for k,v in pairs(eventHandlers[name]) do
    if v == callback then
      return true
    end
  end
  return false
end
--- Gets the event id for the supplied callback
-- @param #string name The event of the callback
-- @param #function callback The callback to request the unique Id of
-- @return #number the unique ID
local function getEventId(name, callback)
  if not eventHandlers[name] then return -1 end
  for k,v in pairs(eventHandlers[name]) do
    if v == callback then
      return tonumber(k)
    end
  end
  return -1
end

--- Register an event 
-- @param #string name The event to register, * to register to all events
-- @param #function callback The function to call when an event hits
-- @return #number the unique id of this event
function BestInSlot:RegisterEvent(name, callback)
  eventId = eventId + 1
  if type(callback) ~= "function" then error("Callback should be a function!") end
  if not eventHandlers[name] then
    eventHandlers[name] = {}
  end
  if not callbackExists(name, callback) then
    eventHandlers[name][tostring(eventId)] = callback
  else
    return getEventId(name, callback)
  end
  return eventId
end

--- Send an event
-- @param #string name The name of the event to send
-- @param #multiple ... Any extra parameters will be forwarded to the eventhandler
-- 
function BestInSlot:SendEvent(name, ...)
  if eventHandlers[name] then
    for _,v in pairs(eventHandlers[name]) do
      v(name, ...)
    end
  end
  if eventHandlers["*"] then
    for _,v in pairs(eventHandlers["*"]) do
      v(name, ...)
    end
  end
end

--- Unregister an eventhandler
-- @param #string name The name of the event to unregister
-- @param #function callback The function to unregister
-- @param #number callback The number to unregister
-- @return #boolean True if successfully removed, false if not
function BestInSlot:UnregisterEvent(name, callback)
  local result = false
  if type(callback) == "function" then
    if eventHandlers[name] then
      for k,v in pairs(eventHandlers[name]) do
        if v == callback then
          eventHandlers[name][k] = nil
          result = true
        end
      end
    end
  elseif type(callback) == "number" then
    for k,v in pairs(eventHandlers) do
      for k2, _ in pairs(v) do
        if k2 == tostring(callback) then
          eventHandlers[k][k2] = nil
          return true
        end
      end
    end
  end
  return result
end