local BestInSlot = LibStub("AceAddon-3.0"):GetAddon("BestInSlot")
local AceGUI = LibStub("AceGUI-3.0")
local L = LibStub("AceLocale-3.0"):GetLocale("BestInSlot")
local window
local isValidText
local validColor = "|cff00ff00"
local invalidColor = "|cffff0000"
local BestInSlotDialog = {}
local dialogsShown = {}
local emptyList = {empty = ""}
local itemId
local raidTier
local sourceType
local instance
local bossId
local difficulty
local OnDone_Func
local isDone
local itemAdded
local eventId
local tryAgainTimerId
--TODO: Add back functionality
local function disableDropdown(widget, selected)
  widget:SetDisabled(true)
  for k,v in pairs(widget.list) do
    if k ~= selected then
      widget:SetItemDisabled(k, true)
    end
  end
end

function BestInSlotDialog:CloseWindow()
  if window ~= nil then
    window:Release()
    window:Hide()
    window = nil
    isValidText = nil
    wipe(dialogsShown)
    sourceType = nil
    itemId = nil
    raidTier = nil
    instance = nil
    bossId = nil
    difficulty = nil
    if OnDone_Func then
      OnDone_Func(itemAdded)
    end
    OnDone_Func = nil
    isDone = false
    itemAdded = false
    if eventId then
      BestInSlot:UnregisterEvent("InitializeLootTable", eventId)
      eventId = nil
    end      
  end
end

function BestInSlotDialog:GetItemId(str)
  local number = tonumber(str)
  if number then
    return number
  end
  local _,link = GetItemInfo(str)
  if not link then return end
  number = (link):match("item:(%d+)")
  return tonumber(number)
end

function BestInSlotDialog:ParseInput(str)
  if #str == 0 then 
    isValidText:SetText("")
    return false 
  end
  local _,link, quality, _, _, _, _, _, equipSlot = GetItemInfo(str)
  if quality then
    if quality >= 4 then
      if not IsEquippableItem(str) then
        isValidText:SetText(invalidColor..L["You can't equip this item!"].."|r")
        return false
      elseif equipSlot == "INVTYPE_BODY" then
        isValidText:SetText(invalidColor..L["You cannot add shirts!"].."|r")
        return false
      elseif equipSlot == "INVTYPE_TABARD" then
        isValidText:SetText(invalidColor..L["You cannot add tabards!"].."|r")
        return false
      end
      isValidText:SetText(validColor..L["This item is valid"].."|r")
      return true
    else
      isValidText:SetText(invalidColor..L["This item is invalid"]..": "..L["The item must be of epic quality or higher"].."|r")
      return false
    end
  else
    isValidText:SetText(invalidColor..(L["This item is invalid"].."! "..L["Note you can find item ID's on %s!"]):format("www.wowhead.com").."|r")
    return false
  end
end

local function itemEditBox_OnTextChanged(widget,_,text, fromScript)
  local result
  if BestInSlotDialog:ParseInput(text) then
    result = false
    widget:DisableButton(result)
    widget:SetUserData("disabled", result)
  else
    result = true
    widget:DisableButton(result)
    widget:SetUserData("disabled", result)
    if not fromScript then
      if tryAgainTimerId then 
        BestInSlot:CancelTimer(tryAgainTimerId)
        tryAgainTimerId = nil
      end
      tryAgainTimerId = BestInSlot:ScheduleTimer(function() 
        if window ~= nil then
          itemEditBox_OnTextChanged(widget, nil, text, true)
          tryAgainTimerId = nil
        end
      end,
      0.25)
    end
  end
  window:DoLayout()
end

local function itemEditBox_OnEnterPressed(widget, _, text)
  if tostring(widget:GetUserData("disabled")) == "false" then
    itemId = BestInSlotDialog:GetItemId(text)
    local _,link = GetItemInfo(itemId)
    widget:SetText(link)
    widget:SetDisabled(true)
    BestInSlotDialog:RaidTierDialog()
  end
end

function BestInSlotDialog:ItemIdDialog()
  if dialogsShown.ItemId then return end
  
  local label = AceGUI:Create("Label")
  label:SetText(BestInSlot.colorNormal..(L["Fill in an %1$sitem ID%2$s or %1$sitem link%2$s in the edit box below"].."|r"):format(BestInSlot.colorHighlight, BestInSlot.colorNormal))
  label:SetFullWidth(true)
  window:AddChild(label)
  local itemEditBox = AceGUI:Create("EditBox")
  itemEditBox:SetFullWidth(true)
  itemEditBox:SetCallback("OnTextChanged", itemEditBox_OnTextChanged)
  itemEditBox:SetCallback("OnEnterPressed",itemEditBox_OnEnterPressed)
  window:AddChild(itemEditBox)
  
  isValidText = AceGUI:Create("Label")
  isValidText:SetFullWidth(true)
  isValidText:SetHeight(26)
  window:AddChild(isValidText)
  itemEditBox_OnTextChanged(itemEditBox, nil, "")
  dialogsShown.ItemId = true
end

local function raidTierDropdown_OnValueChanged(widget, _, value)
  disableDropdown(widget, value)
  widget:SetItemDisabled(value, true)
  raidTier = value
  BestInSlotDialog:SourceDialog()
end

function BestInSlotDialog:RaidTierDialog()
  if dialogsShown.RaidTier then return end
  local label = AceGUI:Create("Label")
  label:SetText(BestInSlot.colorNormal..L["Select the raid tier that you would like to add this item to"].."|r")
  label:SetFullWidth(true)
  window:AddChild(label)
  
  local raidTierDropdown = AceGUI:Create("Dropdown")
  local raidTiers = BestInSlot:GetRaidTiers()
  local dropDown = {}
  for i=1,#raidTiers do
    dropDown[raidTiers[i]] = BestInSlot:GetDescription(BestInSlot.TYPE_RAIDTIER, raidTiers[i])
  end
  raidTierDropdown:SetList(dropDown)
  raidTierDropdown:SetFullWidth(true)
  raidTierDropdown:SetCallback("OnValueChanged", raidTierDropdown_OnValueChanged)
  window:AddChild(raidTierDropdown)
  dialogsShown.RaidTier = true
  if raidTier then
    raidTierDropdown:SetValue(raidTier)
    raidTierDropdown_OnValueChanged(raidTierDropdown, nil, raidTier)
  end
end

local function sourceDropdown_OnValueChanged(widget, _, value)
  disableDropdown(widget, value)
  if value == "MISC" then
    sourceType = "MISC"
    BestInSlotDialog:DifficultyDialog()
  else
    sourceType = "INSTANCE"
    instance = value
    BestInSlotDialog:BossSelectionDialog()
  end
end

function BestInSlotDialog:SourceDialog()
  if dialogsShown.Source then return end
  local label = AceGUI:Create("Label")
  label:SetText(BestInSlot.colorNormal..L["Select the source of the item"].."|r")
  label:SetFullWidth(true)
  window:AddChild(label)
  
  local sourceDropdown = AceGUI:Create("Dropdown")
  local sources = {}
  local order = {}
  local instances = BestInSlot:GetInstances(BestInSlot.TYPE_RAIDTIER, raidTier)
  for i=1,#instances do
    sources[instances[i]] = BestInSlot:GetDescription(BestInSlot.TYPE_INSTANCE, instances[i])
    table.insert(order, instances[i])
  end
  sources.MISC = L["Miscellaneous"]
  table.insert(order, "MISC")
  sourceDropdown:SetList(sources, order)
  sourceDropdown:SetCallback("OnValueChanged", sourceDropdown_OnValueChanged)
  sourceDropdown:SetFullWidth(true)
  window:AddChild(sourceDropdown)
  dialogsShown.Source = true
  if sourceType then
    if sourceType == "MISC" then
      sourceDropdown:SetValue(sourceType)
      sourceDropdown_OnValueChanged(sourceDropdown, nil, sourceType)
    else
      sourceDropdown:SetValue(instance)
      sourceDropdown_OnValueChanged(sourceDropdown, nil, instance)
    end
  end
end

local function bossDropdown_OnValueChanged(widget, _, value)
  disableDropdown(widget, value)
  bossId = value
  BestInSlotDialog:DifficultyDialog()
end

function BestInSlotDialog:BossSelectionDialog()
  local label = AceGUI:Create("Label")
  label:SetText(BestInSlot.colorNormal..(L["Select a boss from %s"]):format(BestInSlot:GetDescription(BestInSlot.TYPE_INSTANCE, instance)).."|r")
  label:SetFullWidth(true)
  window:AddChild(label)
  
  local bossDropdown = AceGUI:Create("Dropdown")
  bossDropdown:SetFullWidth(true)
  local bosses = BestInSlot:GetInstanceBosses(instance)
  bossDropdown:SetList(bosses)
  bossDropdown:SetCallback("OnValueChanged", bossDropdown_OnValueChanged)
  window:AddChild(bossDropdown)
  if bossId then
    bossDropdown:SetValue(bossId)
    bossDropdown_OnValueChanged(bossDropdown, nil, bossId)
  end
end

local function difficDropdown_OnValueChanged(widget,_,value)
  disableDropdown(widget, value)
  difficulty = value
  BestInSlotDialog:FinishDialog()
end

function BestInSlotDialog:DifficultyDialog()
  local label = AceGUI:Create("Label")
  label:SetText(BestInSlot.colorNormal..L["For which difficulty does this item drop?"].."|r")
  label:SetFullWidth(true)
  window:AddChild(label)
  
  local difficDropdown = AceGUI:Create("Dropdown")
  difficDropdown:SetFullWidth(true)
  local difficulties = BestInSlot:GetDifficulties(BestInSlot.TYPE_RAIDTIER, raidTier)
  difficDropdown:SetList(difficulties)
  difficDropdown:SetCallback("OnValueChanged", difficDropdown_OnValueChanged)
  window:AddChild(difficDropdown)
  if difficulty then
    difficDropdown:SetValue(difficulty)
    difficDropdown_OnValueChanged(difficDropdown, nil, difficulty)
  end
end

local function close_OnClick(widget)
  BestInSlotDialog:CloseWindow()
end

local function afterInit()
  local item = BestInSlot:GetItem(itemId)
  if item then
    local label = AceGUI:Create("Label")
    label:SetText(invalidColor..(L["This item allready exists! It's from %s"]):format(item.source))
    label:SetFullWidth(true)
    window:AddChild(label)
    itemAdded = false
  else
    BestInSlot:AddCustomItem(itemId, sourceType, raidTier, difficulty, instance, bossId)
    local label = AceGUI:Create("Label")
    label:SetText(validColor..L["Successfully added the item!"])
    window:AddChild(label)
    itemAdded = true
  end
  local interactiveLabel = BestInSlot:GetItemLinkLabel(itemId)
  interactiveLabel:SetFullWidth(true)
  window:AddChild(interactiveLabel)
  local close = AceGUI:Create("Button")
  close:SetText(L["Close"])
  close:SetFullWidth(true)
  close:SetCallback("OnClick", close_OnClick)
  window:AddChild(close)
  isDone = true
  if eventId then
    BestInSlot:UnregisterEvent("InitializeLootTable", eventId)
    eventId = nil
  end
end

local function submit_OnClick(widget)
  widget:SetDisabled(true)
  if BestInSlot:IsRaidTierInitialized(raidTier) then
    afterInit()
  else
    eventId = BestInSlot:RegisterEvent("InitializeLootTable", afterInit)
    BestInSlot:InitializeLootTable(raidTier)
  end
end
function BestInSlotDialog:FinishDialog()
  local button = AceGUI:Create("Button")
  button:SetText(L["Add item"])
  button:SetCallback("OnClick", submit_OnClick)
  button:SetFullWidth(true)
  window:AddChild(button)
end

function BestInSlot:GetItemDialog(func, args)
  if window then return end
  if type(func) == "function" then
    OnDone_Func = func
  end
  window = AceGUI:Create("Window")
  window:SetTitle(L["Add item"])
  window:EnableResize(nil)
  window:SetWidth(300)
  window:SetHeight(400)
  window:SetLayout("Flow")
  window:SetCallback("OnClose", function() BestInSlotDialog:CloseWindow() end)
  if args then
    raidTier = args.raidTier
    if args.source ~= nil then
      if args.source == "MISC" then
        sourceType = "MISC"
      else
        sourceType = "INSTANCE"
        instance = args.source
      end
    end
    bossId = args.bossId
    difficulty = args.difficulty
  end
  BestInSlotDialog:ItemIdDialog()
end