local BestInSlot = LibStub("AceAddon-3.0"):GetAddon("BestInSlot")
local AceGUI = LibStub("AceGUI-3.0")
local L = LibStub("AceLocale-3.0"):GetLocale("BestInSlot")

local groups = {}
local scrollContainer

--TODO: Implement tickboxes to delete certain ranks from the guild
local function onClickSelectAll(groupContainer, value)
  for i=1,#groupContainer.checkBoxes do
    groupContainer.checkBoxes[i]:SetValue(value)
  end
end

local function getNoGroupLabel()
  local label = AceGUI:Create("Label")
  label:SetRelativeWidth(1)
  label:SetPoint("TOPLEFT", 10, 10)
  label:SetFont(GameFontNormalSmall:GetFont(), 16, nil)
  label:SetText(L["There is no data saved from other players"])
  return label
end

local function deleteSelected()
  local deleteGroups = {}
  for i=1,#groups do
    local deleteCheckBoxes = {}
    for j=1,#groups[i].checkBoxes do
      local checkBox = groups[i].checkBoxes[j]
      if checkBox:GetValue() and checkBox:IsShown() then
        local raidTier = checkBox:GetUserData("raidTier")
        local difficulty = checkBox:GetUserData("difficulty")
        local player = checkBox:GetUserData("name")
        local guild = checkBox:GetUserData("guild")
        BestInSlot.db.factionrealm[guild][player][raidTier][difficulty] = nil
        local result
        table.insert(deleteCheckBoxes, j)
        for k = 1,#groups[i].children do
          if groups[i].children[k] == checkBox then
            result = k
            break
          end
        end
        if result then
          table.remove(groups[i].children, result)
          checkBox:Release()
          groups[i]:DoLayout()
        end          
      end
    end
    table.sort(deleteCheckBoxes, function(a,b) return a > b end)
    for j=1,#deleteCheckBoxes do
      table.remove(groups[i].checkBoxes, deleteCheckBoxes[j])
    end
    if #groups[i].checkBoxes == 0 then
      local result
      for j=1,#scrollContainer.children do
        if scrollContainer.children[j] == groups[i] then
          result = j
          break
        end
      end
      if result then
        table.remove(scrollContainer.children, result)
        groups[i]:Release()
        table.insert(deleteGroups, i)
        scrollContainer:DoLayout()
      end
    end
  end
  table.sort(deleteGroups, function(a,b) return a > b end)
  for i=1,#deleteGroups do
    table.remove(groups, deleteGroups[i])
  end
  if #groups == 0 then
    scrollContainer:ReleaseChildren()
    local label = getNoGroupLabel()
    scrollContainer:AddChild(label)
  end
end

local function triStateHandler(checkBox, checkBoxes)
  local bool = checkBoxes[1]:GetValue()
  for i=2,#checkBoxes do
    if bool ~= checkBoxes[i]:GetValue() then
      checkBox:SetValue(nil)
      return
    end
  end
  checkBox:SetValue(bool)
end

local function onOpen(container)
  container:SetLayout("Fill")
  container:ResumeLayout()
  scrollContainer = AceGUI:Create("ScrollFrame")
  container:AddChild(scrollContainer)
  local cache, raidTiers, guilds = BestInSlot:GetCacheData()
  for i=1,#raidTiers do
    local difficulties = BestInSlot:GetDifficulties(BestInSlot.TYPE_RAIDTIER, raidTiers[i])
    for d=1,#difficulties do
      local difficultyGroup = AceGUI:Create("SimpleGroup")
      difficultyGroup:SetLayout("List")
      difficultyGroup.SelectAll = onClickSelectAll
      difficultyGroup:SetFullWidth(true)
      difficultyGroup.checkBoxes = {}
      local first = true
      local header = AceGUI:Create("Heading")
      header:SetRelativeWidth(1.0)
      header:SetText(("%s - %s"):format(BestInSlot:GetDescription(BestInSlot.TYPE_RAIDTIER, raidTiers[i]), difficulties[d]))
      difficultyGroup.header = header
      local selectAll = AceGUI:Create("CheckBox")
      selectAll:SetLabel(L["Select all"])
      selectAll.group = difficultyGroup
      selectAll:SetTriState(true)
      selectAll:SetCallback("OnValueChanged", function(self,_,value)
        self.group.SelectAll(self.group, value)
      end)
      difficultyGroup.selectAll = selectAll
      for player, savedRaidTiers in pairs(cache) do
        for j=1,#savedRaidTiers do
          if savedRaidTiers[j][1] == raidTiers[i] and d == savedRaidTiers[j][2] then
            if first then
              first = false
              difficultyGroup:AddChild(header)
              difficultyGroup:AddChild(selectAll)
              scrollContainer:AddChild(difficultyGroup)
              table.insert(groups, difficultyGroup)
            end
            local checkBox = AceGUI:Create("CheckBox")
            checkBox:SetLabel(BestInSlot:GetPlayerString(player))
            checkBox.group = difficultyGroup
            checkBox:SetPoint("LEFT", 10, 0)
            checkBox:SetUserData("raidTier", raidTiers[i])
            checkBox:SetUserData("difficulty", d)
            checkBox:SetUserData("name", player)
            checkBox:SetUserData("guild", cache[player].guild)
            checkBox:SetCallback("OnValueChanged", function(self, _, value)
              triStateHandler(self.group.selectAll, self.group.checkBoxes)
            end)
            table.insert(difficultyGroup.checkBoxes, checkBox)
            difficultyGroup:AddChild(checkBox)
          end
        end
      end
    end
  end
  if #groups == 0 then
    local label = getNoGroupLabel()
    scrollContainer:AddChild(label)
  else
    local deleteButton = AceGUI:Create("Button")
    deleteButton:SetText(L["Delete selected data"])
    deleteButton:SetCallback("OnClick", deleteSelected)
    scrollContainer.delete = deleteButton
    scrollContainer:AddChild(deleteButton)
  end
end

local function onClose()
  if scrollContainer then
    scrollContainer:ReleaseChildren()
  end
  wipe(groups)
  scrollContainer = nil
end

BestInSlot:RegisterMenuItem({
  descr = L["Saved BiS Lists"],
  draw = onOpen,
  width = 600,
  height = 600,
  close = onClose,
})