local BestInSlot = LibStub("AceAddon-3.0"):GetAddon("BestInSlot")
local AceGUI = LibStub("AceGUI-3.0")
local L = LibStub("AceLocale-3.0"):GetLocale("BestInSlot")
local frame
local tinsert, tsort = table.insert, table.sort
local menuItems = {}
local container, menuContainer
local selectedMenuItem
local selectedLabel
local selectedMenuId
local MINIMUMWIDTH, MINIMUMHEIGHT = 250, 300
local updateTipsFrame
local tipsIndex = 1
local animating = false
local biggestWidth, biggestHeight = MINIMUMWIDTH, MINIMUMHEIGHT
--Tips to show in the bottom bar
local tips = {
  (L["Developed by %s"]):format("Argathom @ Defias Brotherhood-EU"),
  L["In the BiS manager you can compare items with your selected item!"],
  {
    L["Wanna know how to use the manager?"].." "..L["Put your mouse over this text!"],
    L["You can use this window to edit your BestInSlot list."],
    L["The data in the following dropdown menus are defined by the modules you have activated."],
    L["You can either select an icon or a 'Select an Item' button to select an item for that slot."],
    L["You can shift-click icons or items to link them in the chat."],
    L["After selecting an item, hovering over a different item will popup a compare window."],
  },
  L["Don't like the animating screens? Disable the animations in the options!"],
  (L["Miss a feature that you would like to see? Suggest it on %s"]):format("http://www.curse.com/addons/wow/bestinslot"),
  (L["Do you want this add-on to be available in your language? Help translate at %s"]):format("http://wow.curseforge.com/addons/bestinslot/localization/"),
  (L["%1$s translation by %2$s"]):format(DEDE, "Simalaya @ Defias Brotherhood-EU"),
}

--- Helper for starting an animation
local function startAnimating()
  animating = true
end

--- Helper for stopping an animation
local function stopAnimating()
  animating = false
end

local function onStatusTimer()
  tipsIndex = tipsIndex + 1
  if tipsIndex > #tips then
    tipsIndex = 1
  end
  local newTip = tips[tipsIndex]
  if type(newTip) == "string" then
    frame:SetStatusText(newTip)
  elseif type(newTip) == "table" and type(newTip[1]) == "string" then
    frame:SetStatusText(newTip[1])
  end
end

--- Set the size of the frame
-- @param #number targetWidth The new width
-- @param #number targetHeight the new height
-- @param #function func The function to run when the frame is done resizing
local function SetFrameSize(targetWidth, targetHeight, func)
  if not frame then return end
  if not updateTipsFrame then
    updateTipsFrame = BestInSlot:ScheduleRepeatingTimer(onStatusTimer, 15)
  end
  if BestInSlot.options.windowFixed then
    frame:SetWidth(biggestWidth)
    frame:SetHeight(biggestHeight)
    if func and type(func) == "function" then
      func()
    end
    return
  end
  local max, min = math.max, math.min
  local currentWidth, currentHeight = frame.frame.width, frame.frame.height
  if not targetWidth then targetWidth = currentWidth end
  if not targetHeight then targetHeight = currentHeight end
  if targetWidth == currentWidth and targetHeight == currentHeight then
    if type(func) == "function" then
      func()
    end
    return
  end
  if BestInSlot.options.instantAnimation then
    frame:SetHeight(targetHeight)
    frame:SetWidth(targetWidth)
    if type(func) == "function" then
      func()
    end
  else
    local ANIMATIONTIME = 500 --milliseconds
    local timer = 0
    local incWidth = (targetWidth - currentWidth) / (ANIMATIONTIME / 80)
    local incHeight = (targetHeight - currentHeight) / (ANIMATIONTIME / 80)
    startAnimating()
    local resizeTimerId
    resizeTimerId = BestInSlot:ScheduleRepeatingTimer(function()
      if not frame then stopAnimating() BestInSlot:CancelTimer(resizeTimerId) end
      if incHeight < 0 then
        frame:SetHeight(max(frame.frame.height + incHeight, targetHeight))
      else
        frame:SetHeight(min(frame.frame.height + incHeight, targetHeight))
      end
      if incWidth < 0 then
        frame:SetWidth(max(frame.frame.width + incWidth, targetWidth))
      else
        frame:SetWidth(min(frame.frame.width + incWidth, targetWidth))
      end
      if frame.frame.width == targetWidth and frame.frame.height == targetHeight then
        BestInSlot:CancelTimer(resizeTimerId)
        if func and type(func) == "function" then
          stopAnimating()
          func()
        end
      end
    end,
    0.025)
  end
end

--- Sets the content to the supplied menuWidget
-- @param #table menuWidget An AceGUI Widget with the UserData "id" set to the id of the menu to set the content to
local function SetContent(menuWidget)
  local id = menuWidget:GetUserData("id")
  if not id then error("The menu widget that has been made didn't provide an id") end
  if animating then return false end
  local item = menuItems[id]
  if selectedMenuId then
    local selectedItem = menuItems[selectedMenuId]
    if selectedItem.close then 
      BestInSlot.console:Add("Closing down menu "..selectedItem.descr)
      selectedItem.close() 
    end
    BestInSlot:HideItemTooltip();
  end
  container:ReleaseChildren()
  if item.width then
    SetFrameSize(item.width, item.height, function() container:ResumeLayout() item.draw(container) end)
  else
    item.draw(container)
  end
  BestInSlot.console:Add("Opened menu: "..item.descr)
  selectedMenuId = id
  return true
end
---
-- Requires a table with information as previewed below.
-- Everything between [] is optional
-- {
--  descr = "Description",
--  draw = drawFunction(),
--  [width = 100,]
--  [height = 100,] 
--  [child = "ParentDescription"],
--  [close = closeFunction(),]
-- }
---
function BestInSlot:RegisterMenuItem(parameters)
  if type(parameters) ~= "table" then error("Parameters of RegisterMenuItem should be as described in the documentation") end
  local descr = parameters.descr
  local draw = parameters.draw
  local width = parameters.width
  local height = parameters.height
  local child = parameters.child
  local close = parameters.close
  if not descr or not draw then error("This function should atleast provide a description and a function!") end
  if type(descr) ~= "string" then error("Description parameter should be a string") end
  if type(draw) ~= "function" then error("Draw parameter should be a function") end
  if width and type(width) ~= "number" then error("Width parameter should be a number or nil, param given: "..type(width)) end
  if height and type(height) ~= "number" then error("Height parameter should be a number or nil, param given: "..type(height)) end
  if child and type(child) ~= "string" then error("Child parameter should be a string representing the parents description, or nil, param given: "..type(child)) end
  if close and type(close) ~= "function" then error("Close parameter should be a function, or nil, param given: "..type(close)) end
  if biggestWidth < width then
    biggestWidth = width
  end
  if biggestHeight < height then
    biggestHeight = height
  end
  tinsert(menuItems, parameters)
  return #menuItems
end

function BestInSlot:HideItemTooltip()
  if GameTooltip:IsShown() then
    GameTooltip:Hide()
  end
  if BISComparisonTooltip then
    BISComparisonTooltip:Hide()
  end
end

function BestInSlot:GetItemTooltip(itemid, frame)
  local item = self:GetItem(itemid)
  if not item then return end
  if frame then
    GameTooltip:SetOwner(frame.frame)
    GameTooltip:SetAnchorType("ANCHOR_NONE")
    GameTooltip:SetPoint("TOPLEFT", frame.frame, "TOPRIGHT", 10, 0)
  else
    GameTooltip:SetOwner(UIParent)
  end
  GameTooltip:SetItemByID(itemid)
  GameTooltip.itemid = itemid
  GameTooltip:AddLine("Item ID: "..itemid)
  if item.source then
    GameTooltip:AddLine("Source: "..item.source)
  end
  if BestInSlot.options.DEBUG then
    GameTooltip:AddLine("raidId: "..tostring(item.raidId))
    GameTooltip:AddLine("expansionId: "..tostring(item.expansionId))
    GameTooltip:AddLine("raidTier: "..tostring(item.raidTier))
    GameTooltip:AddLine("link: "..tostring(item.link))
    GameTooltip:AddLine("class: "..tostring(item.class))
    GameTooltip:AddLine("subclass: "..tostring(item.subclass))
    GameTooltip:AddLine("equipSlot: "..tostring(item.equipSlot))
    GameTooltip:AddLine("bossid: "..tostring(item.bossid))
    GameTooltip:AddLine("classRestriction: "..tostring(item.classRestriction))
    GameTooltip:AddLine("difficultyId: "..tostring(item.difficultyId))
    GameTooltip:AddLine("quality: "..tostring(item.quality))
  end
  GameTooltip:Show()
  return GameTooltip
end

function BestInSlot:GetItemComparisonTooltip(tooltip, selectedItem)
  local compareItem = tooltip.itemid
  if not compareItem then return end
  if compareItem == selectedItem then return end --don't show then it's comparing itself!
  if not BISComparisonTooltip or not BISComparisonTooltip:IsShown() then
    CreateFrame("GameTooltip", "BISComparisonTooltip", tooltip, "GameTooltipTemplate")
    BISComparisonTooltip:SetOwner(BestInSlot.frame.frame)
    BISComparisonTooltip:SetAnchorType("ANCHOR_NONE")
    BISComparisonTooltip:SetPoint("TOPLEFT", GameTooltip, "TOPRIGHT")
  end
  BISComparisonTooltip:SetItemByID(selectedItem)
  local _,selectedItemLink = GetItemInfo(selectedItem)
  local _,compareItemLink = GetItemInfo(compareItem)
  local statDelta = GetItemStatDelta(selectedItemLink, compareItemLink)
  for stat,value in pairs(statDelta) do
    local color = ""
    if(value < 0) then
      color = "|cffff0000"
    elseif(value > 0) then
      color = "|cff00ff00"
    end
    BISComparisonTooltip:AddLine(("%s%d %s|r"):format(color, value, tostring(_G[stat])))
  end
  BISComparisonTooltip:Show()
end

function BestInSlot:GameTooltip_OnTooltipSetItem(tooltip, ...)
  if self.db.char.options.showBiSTooltip then
    local _, link = tooltip:GetItem()
    if link then 
      local itemId = (link):match(":%d+:")
      itemId = tonumber((itemId):sub(2, #itemId - 1))
      if itemId then
        local specId = self:IsItemBestInSlot(itemId)
        if specId then
          local _, name = GetSpecializationInfoByID(specId)
          tooltip:AddLine((L["This item is your %1$s (%2$s)"]):format(self.colorHighlight.."BestInSlot"..self.colorNormal, self.colorHighlight..name.."|r"))
        end
      end
    end
  end
end

function BestInSlot:SetTipIndex(id)
  if id > #tips then return end
  if not frame then return end
  tipsIndex = id
  if type(tips[tipsIndex]) == "table" then
    frame:SetStatusText(tips[tipsIndex][1])
  else
    frame:SetStatusText(tips[tipsIndex])
  end
  if updateTipsFrame then
    self:CancelTimer(updateTipsFrame) --Stop the timer temporary so that the tip shows longer
    updateTipsFrame = self:ScheduleTimer(function() 
      updateTipsFrame = self:ScheduleRepeatingTimer(onStatusTimer, 15)
    end, 45) --Enable it again after 45 seconds (+ the 15 of the new timer makes it appear for 1 minute)
  end
end

local function helperOnEnter(widget, ...)
  local itemid = widget:GetUserData("itemid")
  if not itemid then return end
  BestInSlot:GetItemTooltip(itemid, BestInSlot.frame)
  if IsControlKeyDown() then
    SetCursor("INSPECT_CURSOR")
  else
    ResetCursor()
  end
  
  if widget.callbacks.OnShow then
    for i=1,#widget.callbacks.OnShow do
      widget.callbacks.OnShow[i](widget, ...)
    end
  end
end

local function helperOnLeave(widget, ...)
  ResetCursor()
  BestInSlot:HideItemTooltip()
  if widget.callbacks.OnLeave then
    for i=1,#widget.callbacks.OnLeave do
      widget.callbacks.OnLeave[i](widget, ...) 
    end
  end
end

local function helperOnClick(widget, ...)
  local itemlink = widget:GetUserData("itemlink")
  if not itemlink then return end
  if IsShiftKeyDown() then
    if not ChatEdit_InsertLink(itemlink) then
      ChatFrame_OpenChat(itemlink)
    end
  elseif IsControlKeyDown() then
    DressUpItemLink(itemlink)
  end
  if widget.callbacks.OnClick then
    for i=1,#widget.callbacks.OnClick do
      widget.callbacks.OnClick[i](widget, ...) 
    end
  end
end

local function itemLinkCallback(self, name, func)
  if type(func) ~= "function" then error("2nd parameter should be function") end
  if name == "OnEnter" or name == "OnLeave" or name == "OnClick" then
    if not self.callbacks[name] then
      self.callbacks[name] = {}
    end
    table.insert(self.callbacks[name], func)
  else
    error("Invalid callback method")
  end
end

function BestInSlot:GetItemLinkLabel(itemid, withIcon)
  if type(withIcon) == "nil" then
    withIcon = true
  end
  local label = AceGUI:Create("InteractiveLabel")
  label.AddCallback = itemLinkCallback
  local item = self:GetItem(itemid)
  label:SetUserData("itemid", itemid)
  label:SetCallback("OnEnter", helperOnEnter)
  label:SetCallback("OnLeave", helperOnLeave)
  label:SetCallback("OnClick", helperOnClick)
  if not item then
    label:SetText("Undefined itemid")
  else
    label:SetUserData("itemlink", item.link)
    label:SetText(item.link)
  end
  if withIcon then
    label:SetImage(GetItemIcon(itemid))
  end
  label.callbacks = {}
  return label
end

local function OnEnterStatusBar()
  GameTooltip:SetOwner(frame.frame)
  GameTooltip:SetAnchorType("ANCHOR_NONE")
  GameTooltip:SetPoint("TOPLEFT", frame.statustext, "BOTTOMLEFT", -5, 0)
  local tip = tips[tipsIndex]
  if type(tip) == "string" then
    GameTooltip:AddLine(tips[tipsIndex])
  elseif type(tip) == "table" then
    for i=1,#tip do
      if type(tip[i]) == "string" then
        GameTooltip:AddLine(tip[i])
      end
    end
  end
  GameTooltip:Show()
end

local function OnLeaveStatusBar()
  if GameTooltip:IsVisible() then
    GameTooltip:Hide()
  end
end

local function GetFrame()
  if frame then return end
  frame = AceGUI:Create("Frame")
  BestInSlot.frame = frame
  frame:SetTitle("Best In Slot "..GetAddOnMetadata("BestInSlot", "Version"))
  frame:SetStatusText(tips[tipsIndex])
  if not updateTipsFrame then
    updateTipsFrame = BestInSlot:ScheduleRepeatingTimer(onStatusTimer, 15)
  end
  frame:SetCallback( --Set a different OnClose callback to make sure the addon variable 'frame' is set to nil again
    "OnClose", 
    function(widget) 
      BestInSlot.HideFrame(BestInSlot)
    end
  )
  frame:SetCallback("OnEnterStatusBar", OnEnterStatusBar)
  frame:SetCallback("OnLeaveStatusBar", OnLeaveStatusBar)
  frame:EnableResize(nil)
  local dimensionsSet = false
  local initFunc
  if selectedMenuItem then --GUI has been opened before
    for i=1,#menuItems do
      local item = menuItems[i]
      if item.descr == selectedMenuItem then
        initFunc = item.draw
        if item.height then
          frame:SetHeight(item.height)
          frame:SetWidth(item.width)
          dimensionsSet = true
        end
        break
      end
    end
  end
  if not dimensionsSet then
    if BestInSlot.db.char.options.windowFixed then
      frame:SetHeight(biggestHeight)
      frame:SetWidth(biggestWidth)
    else
      frame:SetHeight(MINIMUMHEIGHT)
      frame:SetWidth(MINIMUMWIDTH)
    end
  end
  frame:PauseLayout()
  frame:SetPoint("TOPLEFT", UIParent, "TOPLEFT", 200, -250)
  
  if not BestInSlot.hasModules then
    local label = AceGUI:Create("Label")
    label:SetText(L["This addon requires atleast 1 expansion module! Please enable one!"])
    label:SetFullWidth(true)
    label:SetPoint("TOPLEFT", frame.frame, "TOPLEFT")
    frame:AddChild(label)
    return
  end
  
  menuContainer = AceGUI:Create("SimpleGroup")
  menuContainer:SetPoint("TOPLEFT", frame.frame, "TOPLEFT", 10, -20)
  menuContainer:SetPoint("BOTTOMRIGHT", frame.frame, "BOTTOMLEFT", 160, 50)
  menuContainer:PauseLayout()
  local previousLabel
  for i=1,#menuItems do
    local item = menuItems[i]
    local interactiveLabel = AceGUI:Create("InteractiveLabel")
    interactiveLabel:SetText(item.descr)
    interactiveLabel:SetWidth(140)
    interactiveLabel:SetFont(GameFontNormalSmall:GetFont(), 14, nil)
    interactiveLabel:SetUserData("name", item.descr)
    interactiveLabel:SetUserData("id", i)
    
    if selectedMenuItem and selectedMenuItem == item.descr then
      selectedLabel = interactiveLabel
      interactiveLabel:SetHighlight("Interface\\QuestFrame\\UI-QuestLogTitleHighlight")
      interactiveLabel:SetColor(1,1,1)
    else
      interactiveLabel:SetColor(0.8941, 0.73725, 0.01961)
      interactiveLabel:SetHighlight("Interface\\QuestFrame\\UI-QuestTitleHighlight")
    end
    interactiveLabel:SetCallback("OnClick", function(widget)
      if not selectedLabel or selectedLabel ~= widget then
        if SetContent(widget) then
          if selectedLabel then
            selectedLabel:SetColor(0.8941, 0.73725, 0.01961)
            selectedLabel:SetHighlight("Interface\\QuestFrame\\UI-QuestTitleHighlight")
          end
          widget:SetColor(1,1,1)
          selectedLabel = widget
          selectedMenuItem = widget:GetUserData("name")
          widget:SetHighlight("Interface\\QuestFrame\\UI-QuestLogTitleHighlight")
        end
      end
    end)
    if not previousLabel then
      interactiveLabel:SetPoint("TOPLEFT", menuContainer.frame, "TOPLEFT", 10, -10)
      interactiveLabel:SetPoint("TOPRIGHT", menuContainer.frame, "TOPRIGHT", -10, -10)
    else
      interactiveLabel:SetPoint("TOPLEFT", previousLabel.frame, "BOTTOMLEFT", 0, -10)
      interactiveLabel:SetPoint("TOPRIGHT", previousLabel.frame, "BOTTOMRIGHT", 0, -10)    
    end
    previousLabel = interactiveLabel
    menuContainer:AddChild(interactiveLabel)
  end
  
  frame:AddChild(menuContainer)
  
  container = AceGUI:Create("SimpleGroup")
  container:SetPoint("TOPRIGHT", frame.frame, "TOPRIGHT", -10, -20)
  container:SetPoint("BOTTOMLEFT", menuContainer.frame, "BOTTOMRIGHT", 5, 0)
  frame:AddChild(container)
  
  if initFunc then
    initFunc(container)
  end
end

function BestInSlot:HideFrame()
  if frame then
    if updateTipsFrame then
      BestInSlot:CancelTimer(updateTipsFrame)
      updateTipsFrame = nil
    end
    if selectedMenuId then
      local item = menuItems[selectedMenuId]
      if item and item.close then
        item.close()
      end
    end
    frame:Release()
    frame = nil
    container = nil
    menuContainer = nil
    PlaySound("igCharacterInfoClose")
  end
end

function BestInSlot:ShowFrame()
  if not frame then
    GetFrame()
    PlaySound("igCharacterInfoOpen")
  end
end

function BestInSlot:ToggleFrame()
  if not frame then 
    self:ShowFrame()
  else 
    self:HideFrame()
  end
end
BestInSlot:RegisterSlashCmd("show", (L["%s or %s"]):format("/bis", "/bestinslot"), BestInSlot.ShowFrame, 1)