local BestInSlot = LibStub("AceAddon-3.0"):GetAddon("BestInSlot")
local AceGUI = LibStub("AceGUI-3.0")
local L = LibStub("AceLocale-3.0"):GetLocale("BestInSlot")
local dropdownRaidtier, dropdownDifficulty, dropdownSpecialization, slotContainer --widgets that are referenced later
local selectedRaidTier, selectedDifficulty, selectedSpecialization
local itemGroups = {}
local popupFrame
local eventId

--TODO: Make the select an item frame either a dropdown, or make the popup smaller in width for lower resolution screens
--TODO: Right click to remove items

local function OnAddItem_Complete(result)
  BestInSlot:ShowFrame()
end

local function AddItem_OnClick(widget)
  local args = {
    raidTier = selectedRaidTier,
    difficulty = selectedDifficulty,
  }
  BestInSlot:HideFrame()
  BestInSlot:GetItemDialog(OnAddItem_Complete, args)
end

local function ShowItemList(widget)
  local data = widget:GetUserDataTable() --{raidTier = 50400, difficulty = 1, slotid = 1}
  data.difficulty = data.difficulty or selectedDifficulty
  data.raidTier = data.raidTier or selectedRaidTier
  if not data.difficulty or not data.raidTier or widget.disabled then return end
  local scroll
  local function setTitle()
    if type(BestInSlot.invSlots[data.slotid]) == "string" then
      popupFrame:SetTitle((L["%1$s from raid tier: %2$s"]):format(_G[BestInSlot.invSlots[data.slotid]], BestInSlot:GetDescription(BestInSlot.TYPE_RAIDTIER, data.raidTier)))
    elseif type(BestInSlot.invSlots[data.slotid]) == "table" then
      popupFrame:SetTitle((L["%1$s from raid tier: %2$s"]):format(_G[BestInSlot.invSlots[data.slotid][1]], BestInSlot:GetDescription(BestInSlot.TYPE_RAIDTIER, data.raidTier)))
    end
  end
  local function iconOnEnter(widget)
    local id = widget:GetUserData("id")
    local tooltip = BestInSlot:GetItemTooltip(id, popupFrame)
    local itemid = BestInSlot.db.char[scroll.data.raidTier][scroll.data.difficulty][selectedSpecialization][scroll.data.slotid]
    if type(itemid) == "number" then
      BestInSlot:GetItemComparisonTooltip(tooltip, itemid)
    end 
  end
  local function iconOnClick(widget)
    local itemlink = widget:GetUserData("itemlink")
    local id = widget:GetUserData("id")
    if IsShiftKeyDown() then
      if not ChatEdit_InsertLink(itemlink) then
        ChatFrame_OpenChat(itemlink)
      end
    else
      BestInSlot:SetItemBestInSlot(scroll.data.raidTier, scroll.data.difficulty, selectedSpecialization, scroll.data.slotid, id)
      slotContainer.populateSlots()
      if BISComparisonTooltip then
        BISComparisonTooltip:SetItemByID(id)
      end
    end
  end
  local function iconOnLeave() BestInSlot:HideItemTooltip() end
  if popupFrame then
    setTitle()
    scroll = popupFrame.scroll
    scroll.data = data
    scroll:SetScroll(0)
  else
    popupFrame = AceGUI:Create("Frame")
    popupFrame:EnableResize(false)
    popupFrame:SetWidth(300)
    popupFrame:SetLayout("Fill")
    setTitle()
    popupFrame:SetPoint("TOPLEFT", BestInSlot.frame.frame, "TOPRIGHT")
    popupFrame:SetPoint("BOTTOMLEFT", BestInSlot.frame.frame, "BOTTOMRIGHT")
    popupFrame:SetCallback("OnClose", function() popupFrame:Release() popupFrame = nil end)
    
    local scrollContainer = AceGUI:Create("SimpleGroup")
    scrollContainer:SetFullWidth(true)
    scrollContainer:SetFullHeight(true)
    scrollContainer:SetLayout("Fill")
    popupFrame:AddChild(scrollContainer)
        
    scroll = AceGUI:Create("ScrollFrame")
    popupFrame.scroll = scroll
    scroll.data = data
    scroll:SetLayout("Flow")
    function scroll.SetLootTable(lootTable)
      local ICONHEIGHT = 30
      scroll:ReleaseChildren()
      local icon = AceGUI:Create("Icon")
      icon:SetImage("Interface\\Buttons\\UI-GroupLoot-Pass-Up.blp")
      icon:SetImageSize(ICONHEIGHT,ICONHEIGHT)
      icon:SetHeight(ICONHEIGHT)
      icon:SetLabel(L["Deselect item"])
      icon:SetFullWidth(true)
      icon:SetCallback("OnClick", function(widget, event, button)
        BestInSlot:SetItemBestInSlot(scroll.data.raidTier, scroll.data.difficulty, selectedSpecialization, scroll.data.slotid, nil)
        slotContainer.populateSlots()
      end)
      --itemGroup:AddChild(icon)
      scroll:AddChild(icon)
      local itemOrder = BestInSlot:GetLootOrder(lootTable)
      for i = 1, #itemOrder do
        local id = itemOrder[i]
        local item = lootTable[id]        
        local icon = AceGUI:Create("Icon")
        local texture = GetItemIcon(id)
        icon:SetImage(texture)
        icon:SetImageSize(ICONHEIGHT,ICONHEIGHT)
        icon:SetHeight(ICONHEIGHT)
        icon:SetUserData("id", id)
        icon:SetUserData("itemlink", item.link)
        icon:SetRelativeWidth(0.49)
        icon:SetLabel(item.link)
        icon:SetCallback("OnEnter", iconOnEnter)
        icon:SetCallback("OnLeave", iconOnLeave)
        icon:SetCallback("OnClick", iconOnClick)
        scroll:AddChild(icon)
      end
      local icon = AceGUI:Create("Icon")
      icon:SetImage("Interface\\GuildBankFrame\\UI-GuildBankFrame-NewTab.png")
      icon:SetImageSize(ICONHEIGHT, ICONHEIGHT)
      icon:SetHeight(ICONHEIGHT)
      icon:SetFullWidth(true)
      icon:SetLabel(L["Add item"])
      icon:SetCallback("OnClick", AddItem_OnClick)
      scroll:AddChild(icon)
    end
    scrollContainer:AddChild(scroll)
  end
  scroll.SetLootTable(BestInSlot:GetPersonalizedLootTableBySlot(data.raidTier, data.slotid, data.difficulty, selectedSpecialization))
end

local function GetSlotContainer(raidTier, difficulty)
  local container = AceGUI:Create("SimpleGroup")
  wipe(itemGroups)
  container:SetFullWidth(true)
  container:SetHeight(0)
  container:SetLayout("List")
  local previousGroup
  local slots = BestInSlot.slots
  local function iconOnEnter(icon)
    if icon.itemid then
      if not popupFrame then
        BestInSlot:GetItemTooltip(icon.itemid, BestInSlot.frame)
      elseif popupFrame then
        BestInSlot:GetItemTooltip(icon.itemid, popupFrame)
      end
    end
  end
  local function iconOnLeave() BestInSlot:HideItemTooltip()  end
  local function iconOnClick(widget)
    if IsShiftKeyDown() and widget.itemid then
      if not ChatEdit_InsertLink(BestInSlot:GetItem(widget.itemid).link) then
        ChatFrame_OpenChat(BestInSlot:GetItem(widget.itemid).link)
      end
    elseif not IsShiftKeyDown() then
      ShowItemList(widget)
    end
  end
  for i=1,#slots do
    local slotId, textureName = GetInventorySlotInfo(slots[i])
    local itemGroup = AceGUI:Create("SimpleGroup")
    itemGroup.defaultTexture = textureName
    itemGroup.slotid = slotId
    itemGroup:SetHeight(45)
    itemGroup:PauseLayout()
    
    local icon = AceGUI:Create("Icon")
    icon:SetImage(textureName)
    icon:SetImageSize(40,40)
    icon:SetWidth(40)
    icon:SetHeight(45)
    icon:SetPoint("TOPLEFT", itemGroup.frame, "TOPLEFT")
    icon:SetCallback("OnEnter", iconOnEnter)
    icon:SetCallback("OnLeave", iconOnLeave)
    icon:SetCallback("OnClick", iconOnClick)
    icon:SetUserData("slotid", slotId)
    itemGroup:AddChild(icon)
    itemGroup.icon = icon
    
    local label = AceGUI:Create("Label")
    label:SetHeight(20)
    label:SetPoint("TOPLEFT", icon.frame, "TOPRIGHT", 0, -5)
    itemGroup:AddChild(label)
    itemGroup.label = label
    
    local button = AceGUI:Create("Button")
    button:SetText(L["Select an item"])
    button:SetCallback("OnClick", ShowItemList)
    button:SetHeight(25)
    button:SetPoint("BOTTOMLEFT", icon.frame, "BOTTOMRIGHT")
    button:SetUserData("slotid", slotId)
    itemGroup:AddChild(button)
    itemGroup.button = button    
    
    if i == 7 then 
      container:PauseLayout() 
      itemGroup:SetPoint("TOPRIGHT", container.frame, "TOPRIGHT")
      previousGroup = itemGroup
    end
    if i > 7 and i < 15 then
      itemGroup:SetPoint("TOPRIGHT", previousGroup.frame, "BOTTOMRIGHT")
      previousGroup = itemGroup
    elseif i == 15 then
      itemGroup:SetPoint("BOTTOMRIGHT", container.frame, "BOTTOM", 0, 10)
    elseif i == 16 then
      itemGroup:SetPoint("BOTTOMLEFT", container.frame, "BOTTOM", 0, 10)
    end
    container:AddChild(itemGroup)
    tinsert(itemGroups, itemGroup)
  end
  function container.populateSlots()
    local container = slotContainer
    local isTwoHander = false
    for i=1,#itemGroups do
      local itemGroup = itemGroups[i]
      local itemid = BestInSlot.db.char[selectedRaidTier][selectedDifficulty][selectedSpecialization][itemGroup.slotid]
      --Code for disabling the off-hand slot in Two-handed cases
      if itemGroup.slotid == 17 and isTwoHander then
        BestInSlot:SetItemBestInSlot(selectedRaidTier, selectedDifficulty, selectedSpecialization, itemGroup.slotid, nil)
        itemGroup.icon.itemid = nil
        itemGroup.icon.disabled = true
        
        itemGroup.button:SetDisabled(true)
        
        itemid = nil
      else
        itemGroup.icon.disabled = false
        
        itemGroup.button:SetDisabled(false)
      end
      if type(itemid) == "number" and BestInSlot:GetItem(itemid) then --Has a valid itemid value in the database
        local item = BestInSlot:GetItem(itemid)
        if itemGroup.slotid == 16 then
          if item.equipSlot == "INVTYPE_2HWEAPON" or item.equipSlot == "INVTYPE_RANGED" or item.equipSlot == "INVTYPE_RANGEDRIGHT" then
            isTwoHander = true
          end
        end
        itemGroup.icon:SetImage(GetItemIcon(itemid))
        itemGroup.icon.itemid = itemid
        
        itemGroup.label:SetText(item.link)
        
        itemGroup.button:SetDisabled(false)
      else
        itemGroup.icon:SetImage(itemGroup.defaultTexture)
        itemGroup.icon.itemid = nil
        
        itemGroup.label:SetText("")
      end
      itemGroup.icon:SetUserData("raidTier", selectedRaidTier)
      itemGroup.icon:SetUserData("difficulty", selectedDifficulty)
      itemGroup.button:SetUserData("raidTier", selectedRaidTier)
      itemGroup.button:SetUserData("difficulty", selectedDifficulty)
    end
    if eventId then
      BestInSlot:UnregisterEvent("InitializeLootTable", eventId)
      eventId = nil
    end
  end
  return container
end

local function GetDropdownRaidTier()
  local options = {}
  local newSelection
  local selection = selectedRaidTier or BestInSlot:GetLatest(BestInSlot.TYPE_RAIDTIER)
  local raidTiers = BestInSlot:GetRaidTiers()
  for i=1,#raidTiers do
    options[raidTiers[i]] = BestInSlot:GetDescription(BestInSlot.TYPE_RAIDTIER, raidTiers[i])
  end
  return options, selection
end

local function GetBestInSlotManager(container)
  container:PauseLayout()
  
  dropdownRaidtier = AceGUI:Create("Dropdown")
  dropdownRaidtier:SetLabel(L["Raid Tier"])
  dropdownRaidtier:SetPoint("TOPLEFT", container.frame, "TOPLEFT", 10, -10)
  local options, selection = GetDropdownRaidTier()
  selectedRaidTier = selection
  dropdownRaidtier:SetList(options)
  dropdownRaidtier:SetValue(selection)
  dropdownRaidtier:SetCallback("OnValueChanged",
    function(_,_,value)
      selectedRaidTier = value
      local difficulties = BestInSlot:GetDifficulties(BestInSlot.TYPE_RAIDTIER, value)
      dropdownDifficulty:SetList(difficulties)
      dropdownDifficulty:SetValue(#difficulties)
      selectedDifficulty = #difficulties;
      if BestInSlot:IsRaidTierInitialized(value) then
        slotContainer.populateSlots()
      else
        eventId = BestInSlot:RegisterEvent("InitializeLootTable", slotContainer.populateSlots)
        BestInSlot:InitializeLootTable(value)
      end
      if popupFrame ~= nil then
        popupFrame:Release()
        popupFrame = nil
      end
    end
  )
  container:AddChild(dropdownRaidtier)
  
  dropdownDifficulty = AceGUI:Create("Dropdown")

  dropdownDifficulty:SetLabel(L["Difficulty"])
  local difficulties = BestInSlot:GetDifficulties(BestInSlot.TYPE_RAIDTIER, selectedRaidTier)
  if not selectedDifficulty then
    selectedDifficulty = #difficulties
  end
  dropdownDifficulty:SetList(difficulties)
  dropdownDifficulty:SetPoint("TOPLEFT", dropdownRaidtier.frame, "TOPRIGHT")
  dropdownDifficulty:SetValue(selectedDifficulty)
  dropdownDifficulty:SetCallback("OnValueChanged",
    function(_,_,value)
      selectedDifficulty = value
      slotContainer.populateSlots()
      if popupFrame ~= nil then
        popupFrame:Release()
        popupFrame = nil
      end
    end
  )
  container:AddChild(dropdownDifficulty)
  
  dropdownSpecialization = AceGUI:Create("Dropdown")
  local specs = {}
  local selected = selectedSpecialization
  for i=1,GetNumSpecializations() do
    local id, name = GetSpecializationInfo(i)
    specs[id] = name
  end
  if not selected then
    selected = GetSpecializationInfo(BestInSlot:GetSpecialization())
  end
  selectedSpecialization = selected
  dropdownSpecialization:SetList(specs)
  dropdownSpecialization:SetValue(selected)
  dropdownSpecialization:SetPoint("TOPLEFT", dropdownDifficulty.frame, "TOPRIGHT")
  dropdownSpecialization:SetLabel(L["Specialization"])
  dropdownSpecialization:SetCallback("OnValueChanged", 
    function(_,_, value)
      selectedSpecialization = value
      slotContainer.populateSlots()
      if popupFrame and popupFrame.scroll and popupFrame.scroll.SetLootTable then
        popupFrame.scroll.SetLootTable(BestInSlot:GetPersonalizedLootTableBySlot(selectedRaidTier, popupFrame.scroll.data.slotid, selectedDifficulty, selectedSpecialization))
      end
    end
  )
  
  container:AddChild(dropdownSpecialization)
  
  slotContainer = GetSlotContainer(selectedRaidTier,1)
  if BestInSlot:IsRaidTierInitialized(selectedRaidTier) then
    slotContainer.populateSlots()
  else
    eventId = BestInSlot:RegisterEvent("InitializeLootTable", slotContainer.populateSlots)
    BestInSlot:InitializeLootTable(selectedRaidTier)
  end
  slotContainer:SetPoint("TOPLEFT", dropdownRaidtier.frame, "BOTTOMLEFT")
  slotContainer:SetPoint("BOTTOMRIGHT", container.frame, "BOTTOMRIGHT")
  container:AddChild(slotContainer)
  
  BestInSlot:SetTipIndex(3)
end

local function close()
  BestInSlot:HideItemTooltip()
  if popupFrame then
    popupFrame:Release()
    popupFrame = nil
  end
  dropdownRaidtier = nil
  dropdownDifficulty = nil
  dropdownSpecialization = nil
  slotContainer = nil
  wipe(itemGroups)
  if eventId then
    BestInSlot:UnregisterEvent("InitializeLootTable", eventId)
  end
end

BestInSlot:RegisterMenuItem({
  descr = (L["%s manager"]):format("BiS"),
  draw = GetBestInSlotManager,
  width = 800,
  height = 600,
  close = close,
})