local BestInSlot = LibStub("AceAddon-3.0"):GetAddon("BestInSlot")
local AceGUI = LibStub("AceGUI-3.0")
local L = LibStub("AceLocale-3.0"):GetLocale("BestInSlot")


local function drawMenu(container)
  container:SetLayout("Flow")
  local checkBox = AceGUI:Create("CheckBox")
  checkBox:SetLabel(L["Debug messages"])
  checkBox:SetDescription(
    L["Debug messages will be shown in the chat window if enabled."].."\n"..
    L["These can be used to identify problems."].."\n"..
    L["Generally encouraged to keep this disabled"].."\n"..
    L["This setting is not saved between sessions, and off by default."]
  )
  checkBox:SetRelativeWidth(0.5)
  checkBox:SetValue(BestInSlot.options.DEBUG)
  checkBox:SetCallback("OnValueChanged", function(_,_,value) 
    BestInSlot.options.DEBUG = value 
    BestInSlot:SendEvent("DebugOptionsChanged", value) 
  end)
  container:AddChild(checkBox)
  
  checkBox = AceGUI:Create("CheckBox")
  checkBox:SetLabel(L["Instant animations"])
  checkBox:SetDescription(
    L["Changes the behavior of the BestInSlot frame."].."\n"..
    L["Checking this disables the animation, and makes the frame instantly change size."].."\n"..
    L["This setting is account wide and saved between sessions."]
  )
  checkBox:SetRelativeWidth(0.5)
  checkBox:SetValue(BestInSlot.options.instantAnimation)
  checkBox:SetCallback("OnValueChanged", function(_,_,value) 
    BestInSlot.options.instantAnimation = value 
    BestInSlot.db.global.options.instantAnimation = value  
  end)
  container:AddChild(checkBox)
  
  checkBox = AceGUI:Create("CheckBox")
  checkBox:SetLabel(L["Show BestInSlot in Item tooltip"])
  checkBox:SetRelativeWidth(0.5)
  checkBox:SetDescription(
    L["This will put a note underneath Item tooltips when the item is part of your BestInSlot"]
  )
  checkBox:SetValue(BestInSlot.options.showBiSTooltip)
  checkBox:SetCallback("OnValueChanged", function(_,_,value)
    BestInSlot.options.showBiSTooltip = value
    BestInSlot.db.char.options.showBiSTooltip = value  
  end)
  container:AddChild(checkBox)
  
  checkBox = AceGUI:Create("CheckBox")
  checkBox:SetLabel(L["Disable resizing"])
  checkBox:SetRelativeWidth(0.5)
  checkBox:SetDescription(
    L["This will disable resizing and set the window on a fixed size"]
  )
  checkBox:SetValue(BestInSlot.options.windowFixed)
  checkBox:SetCallback("OnValueChanged", function(_,_,value)
    BestInSlot.options.windowFixed = value
    BestInSlot.db.char.options.windowFixed = value
  end)
  container:AddChild(checkBox)
  
  checkBox = AceGUI:Create("CheckBox")
  checkBox:SetLabel(L["Send automatic updates"])
  checkBox:SetRelativeWidth(0.5)
  checkBox:SetDescription(
    L["When you update your BiS list, it'll automatically send the updates to the guild"]
  )
  checkBox:SetValue(BestInSlot.options.sendAutomaticUpdates)
  checkBox:SetCallback("OnValueChanged", function(_,_,value)
    BestInSlot.options.sendAutomaticUpdates = value
    BestInSlot.db.char.options.sendAutomaticUpdates = value
  end)
  container:AddChild(checkBox)
  
  checkBox = AceGUI:Create("CheckBox")
  checkBox:SetLabel(L["Receive automatic updates"])
  checkBox:SetRelativeWidth(0.5)
  checkBox:SetDescription(
    L["When others in your guild update their BiS list, automatically save it."]
  )
  checkBox:SetValue(BestInSlot.options.receiveAutomaticUpdates)
  checkBox:SetCallback("OnValueChanged", function(_,_,value)
    BestInSlot.options.receiveAutomaticUpdates = value
    BestInSlot.db.char.options.receiveAutomaticUpdates = value
  end)
  container:AddChild(checkBox)
  
  checkBox = AceGUI:Create("CheckBox")
  checkBox:SetLabel(L["Show minimap button"])
  checkBox:SetRelativeWidth(0.5)
  checkBox:SetValue(BestInSlot.db.char.options.minimapButton)
  checkBox:SetCallback("OnValueChanged", BestInSlot.MiniMapButtonHideShow)
  container:AddChild(checkBox)
end

local function close()

end

BestInSlot:RegisterMenuItem({
  descr = L["Options"],
  draw = drawMenu,
  width = 800,
  height = 400,
  close = close,
})