local BestInSlot = LibStub("AceAddon-3.0"):GetAddon("BestInSlot")
local AceGUI = LibStub("AceGUI-3.0")
local L = LibStub("AceLocale-3.0"):GetLocale("BestInSlot")
local selectedDungeon, selectedDifficulty, selectedReport
local dropdownDungeon, dropdownDifficulty, dropdownReport, dropdownFilter
local reportButton
local contentContainer
local hasContent
local requiredItems = {}
local eventId
local specFilter = {}

local function getDropdownDungeons(raidTier)
  local options = {}
  local instances = BestInSlot:GetInstances()
  local selected = selectedDungeon
  if not selected then
    local selectedRaidTier
    for i = 1,#instances do
      local raidTier = BestInSlot:GetRaidTiers(BestInSlot.TYPE_INSTANCE, instances[i])
      if not selectedRaidTier or selectedRaidTier < raidTier then
        selected = instances[i]
        selectedRaidTier = raidTier
      end
    end
  end
  for i=1,#instances do
    options[instances[i]] = BestInSlot:GetDescription(BestInSlot.TYPE_INSTANCE, instances[i])
  end
  return options, selected
end

local function getDropdownDifficulties(raidTier)
  local options = BestInSlot:GetDifficulties(BestInSlot.TYPE_RAIDTIER, raidTier)
  local selected = selectedDifficulty or #options or 1
  if selected > #options then
    selected = #options
  end
  return options, selected
end

local function tableSize(table)
  local size = 0
  for _ in pairs(table) do
    size = size + 1
  end
  if size == 0 then size = nil end
  return size
end

local function reportRequiredItems()
  local playerName = GetUnitName("player")
  local notAcquiredItems = {}
  for i=1,#requiredItems do
    if #requiredItems[i] > 0 then
      for j=1,#requiredItems[i] do
        if not BestInSlot:HasItem(requiredItems[i][j]) then
          if not notAcquiredItems[i] then
            notAcquiredItems[i] = {}
          end
          table.insert(notAcquiredItems[i], requiredItems[i][j])
        end
      end
    end
  end
  local str = {}
  for i,list in pairs(notAcquiredItems) do
    local bossName = BestInSlot:GetDescription(BestInSlot.TYPE_BOSS, selectedDungeon, i)
    str[#str+1] = bossName
    for j=1,#list do
      if (#list[j] + #str[#str]) > 250 then
        str[#str + 1] = bossName..": "..list[j]
      else
        if j == 1 then
          str[#str] = str[#str]..": "..list[j]
        else
          str[#str] = str[#str].." & "..list[j]
        end
      end
    end
  end
  if not tableSize(notAcquiredItems) then
    SendChatMessage(
      (L["%1$s needs nothing from %2$s %3$s"]):format(
        playerName,
        BestInSlot:GetDescription(BestInSlot.TYPE_INSTANCE, selectedDungeon),
        BestInSlot:GetDifficulties(BestInSlot.TYPE_INSTANCE, selectedDungeon)[selectedDifficulty]
      ),
      selectedReport
    )
  else
    SendChatMessage(
      (L["%1$s needs the following bosses from %2$s %3$s"]):format(
        playerName, 
        BestInSlot:GetDescription(BestInSlot.TYPE_INSTANCE, selectedDungeon), 
        BestInSlot:GetDifficulties(BestInSlot.TYPE_INSTANCE, selectedDungeon)[selectedDifficulty]
      ), 
      selectedReport
    )
    for i=1,#str do
      SendChatMessage(str[i], selectedReport)
    end
  end
end

local function count(table)
  local counter = 0
  for _, v in pairs(table) do
    if v then
      counter = counter + 1
    end
  end
  return counter
end

local function updateContent()
  wipe(requiredItems)
  contentContainer:ReleaseChildren()
  local bossList = BestInSlot:GetInstanceBosses(selectedDungeon)
  local bisList = BestInSlot.db.char[BestInSlot:GetRaidTiers(BestInSlot.TYPE_INSTANCE, selectedDungeon)][selectedDifficulty]
  local difficulties = BestInSlot:GetDifficulties(BestInSlot.TYPE_INSTANCE, selectedDungeon)
  local bossLootGroups = {}
  for i=1,#bossList do
    local bossGroup = AceGUI:Create("SimpleGroup")
    bossGroup:SetFullWidth(true)
    bossGroup:SetLayout("Flow")
    local label = AceGUI:Create("Heading")
    label:SetFullWidth(true)
    label:SetText(bossList[i])
    bossGroup:AddChild(label)
    local hasItems = false
    table.insert(bossLootGroups, bossGroup)
    requiredItems[i] = {}
  end
  local slots = BestInSlot.slots
  local specs = BestInSlot:GetAllSpecializations()
  for specId, specName in pairs(specs) do
    if count(specFilter) == 0 or specFilter[specId] then
      for i=1,#slots do
        local slotId = GetInventorySlotInfo(slots[i])
        if type(bisList[specId][slotId]) == "number" then
          local id = bisList[specId][slotId]
          local item = BestInSlot:GetItem(id)
          if item and item.raidId and item.raidId == selectedDungeon and item.bossid then
            local label = BestInSlot:GetItemLinkLabel(id)
            label:SetFullWidth(true)
            label:SetFont(GameFontNormalSmall:GetFont(), 14, nil)
            local text = item.link
            if count(specFilter) ~= 1 then
              text = ("%s (%s)"):format(text, specName)
            end
            if BestInSlot:HasItem(id) then
              text = ("%s - %s"):format(text, L["You have this item"])
            end
            label:SetText(text)
            local bossGroup = bossLootGroups[item.bossid]
            bossGroup:AddChild(label)
            table.insert(requiredItems[item.bossid], item.link)
          end
        end
      end
    end
  end
  for i=#bossLootGroups,1, -1 do
    if #bossLootGroups[i].children == 1 then --When it only has the boss label, thus no loot
      bossLootGroups[i]:Release()
      table.remove(bossLootGroups, i)
    end
  end
  if #bossLootGroups == 0 then
    local label = AceGUI:Create("Label")
    label:SetFullWidth(true)
    label:SetFont(GameFontNormalSmall:GetFont(), 14, nil)
    label:SetText(
      "\n"..L["It looks like this dungeon does not drop any of your BestInSlot items."].."\n"..
      "\n"..L["It could also be that you haven't set the BestInSlot for this raid tier yet"].."\n"..
      "\n"..L["You can use the BiS Manager on the left to set your BestInSlot"]
    )
    hasContent = false
    reportButton:SetDisabled(true)
    contentContainer:AddChild(label)
  else
    hasContent = true
    if selectedReport then
      reportButton:SetDisabled(false)
    end
    for i=1,#bossLootGroups do
      contentContainer:AddChild(bossLootGroups[i])
    end
  end
  if eventId then
    BestInSlot:UnregisterEvent("InitializeLootTable", eventId)
    eventId = nil
  end
end

local function dropdownDungeonOnValueChanged(_,_,value)
  selectedDungeon = value
  local options, selected = getDropdownDifficulties(BestInSlot:GetRaidTiers(BestInSlot.TYPE_INSTANCE, selectedDungeon))
  dropdownDifficulty:SetList(options)
  dropdownDifficulty:SetValue(selected)
  selectedDifficulty = selected
  local raidTier = BestInSlot:GetRaidTiers(BestInSlot.TYPE_INSTANCE, selectedDungeon)
  if not BestInSlot:IsRaidTierInitialized(raidTier) then
    eventId = BestInSlot:RegisterEvent("InitializeLootTable", updateContent)
    BestInSlot:InitializeLootTable(raidTier)
  else
    updateContent()
  end
end

local function dropdownDifficultyOnValueChanged(_,_,value)
  selectedDifficulty = value
  local raidTier = BestInSlot:GetRaidTiers(BestInSlot.TYPE_INSTANCE, selectedDungeon)
  if not BestInSlot:IsRaidTierInitialized(raidTier) then
    eventId = BestInSlot:RegisterEvent("InitializeLootTable", updateContent)
    BestInSlot:InitializeLootTable(raidTier)
  else
    updateContent()
  end
end

local function dropdownFilterOnValueChanged(widget,event, spec, value)
  specFilter[spec] = value
  updateContent()
end

local function MakeContainer(container)
  container:PauseLayout()
  
  dropdownDungeon = AceGUI:Create("Dropdown")
  local options,selected = getDropdownDungeons(selectedRaidTier)
  dropdownDungeon:SetList(options)
  dropdownDungeon:SetValue(selected)
  selectedDungeon = selected
  dropdownDungeon:SetLabel(L["Raid Instance"])
  dropdownDungeon:SetPoint("TOPLEFT", container.frame, "TOPLEFT", 10, -10)
  dropdownDungeon:SetCallback("OnValueChanged", dropdownDungeonOnValueChanged)
  
  container:AddChild(dropdownDungeon)
  
  dropdownDifficulty = AceGUI:Create("Dropdown")
  local options, selected = getDropdownDifficulties(BestInSlot:GetRaidTiers(BestInSlot.TYPE_INSTANCE, selectedDungeon))
  dropdownDifficulty:SetList(options)
  dropdownDifficulty:SetValue(selected)
  dropdownDifficulty:SetLabel(L["Difficulty"])
  dropdownDifficulty:SetPoint("TOPLEFT", dropdownDungeon.frame, "TOPRIGHT")
  dropdownDifficulty:SetWidth(120)
  selectedDifficulty = selected
  
  dropdownDifficulty:SetCallback("OnValueChanged", dropdownDifficultyOnValueChanged)
  
  container:AddChild(dropdownDifficulty)
    
  local specs = BestInSlot:GetAllSpecializations()
  dropdownFilter = AceGUI:Create("Dropdown")
  dropdownFilter:SetList(specs)
  dropdownFilter:SetMultiselect(true)
  if count(specFilter) == 0 then
    local filter = GetSpecializationInfo(BestInSlot:GetSpecialization())
    dropdownFilter:SetItemValue(filter, true)
    specFilter[filter] = true
  else
    for specId in pairs(specs) do
      if specFilter[specId] then
        dropdownFilter:SetItemValue(specId, true)
      end
    end
  end
  dropdownFilter:SetLabel(FILTER) -- Blizzard global string
  dropdownFilter:SetPoint("TOPLEFT", dropdownDifficulty.frame, "TOPRIGHT")
  dropdownFilter:SetPoint("TOPRIGHT", container.frame, "TOPRIGHT", -10, -10)
  dropdownFilter:SetCallback("OnValueChanged", dropdownFilterOnValueChanged)
  
  container:AddChild(dropdownFilter)
  
  contentContainer = AceGUI:Create("ScrollFrame")
  contentContainer:SetPoint("TOPLEFT", dropdownDungeon.frame, "BOTTOMLEFT")
  contentContainer:SetPoint("BOTTOMRIGHT", container.frame, "BOTTOMRIGHT", -10, 50)
  contentContainer:SetLayout("List")
  
  container:AddChild(contentContainer)
  dropdownReport = AceGUI:Create("Dropdown")
  dropdownReport:SetPoint("TOPLEFT", contentContainer.frame, "BOTTOMLEFT")
  dropdownReport:SetLabel(L["Report your BestInSlot to:"])
  dropdownReport:SetText(L["Select an option"])
  local dropdownList = {
    SAY = _G["SAY"],
    PARTY = _G["PARTY"],
    GUILD = _G["GUILD"],
    OFFICER = _G["OFFICER"],
    RAID = _G["RAID"],
    INSTANCE_CHAT = _G["INSTANCE_CHAT"]
  }
  dropdownReport:SetList(dropdownList)
  if not IsInGroup() then
    dropdownReport:SetItemDisabled("PARTY", true)
  end
  if not IsInGuild() then
    dropdownReport:SetItemDisabled("GUILD", true)
    dropdownReport:SetItemDisabled("OFFICER", true)
  end
  --TODO: Check if player is an officer when API provides a function to do so
  if not UnitInRaid("player") then
    dropdownReport:SetItemDisabled("RAID", true)
  end
  if not IsInGroup(LE_PARTY_CATEGORY_INSTANCE) then
    dropdownReport:SetItemDisabled("INSTANCE_CHAT", true)
  end
  if selectedReport then
    dropdownReport:SetValue(selectedReport)
  end
  dropdownReport:SetCallback("OnValueChanged", function(_,_, value)
    selectedReport = value
    if hasContent then
      reportButton:SetDisabled(false)
    end
  end)
  container:AddChild(dropdownReport)
  
  reportButton = AceGUI:Create("Button")
  reportButton:SetText(L["Report"])
  reportButton:SetDisabled(selectedReport == nil)
  reportButton:SetPoint("BOTTOMLEFT", dropdownReport.frame, "BOTTOMRIGHT")
  reportButton:SetCallback("OnClick", reportRequiredItems)
  
  container:AddChild(reportButton)
  local raidTier = BestInSlot:GetRaidTiers(BestInSlot.TYPE_INSTANCE, selectedDungeon)
  if not BestInSlot:IsRaidTierInitialized(raidTier) then
    eventId = BestInSlot:RegisterEvent("InitializeLootTable", updateContent)
    BestInSlot:InitializeLootTable(raidTier)
  else
    updateContent()
  end
end

local function onClose()
  dropdownDungeon, dropdownDifficulty, dropdownReport, dropdownFilter = nil
  reportButton = nil
  contentContainer = nil
  hasContent = nil
  wipe(requiredItems)
  if eventId then
    BestInSlot:UnregisterEvent("InitializeLootTable", eventId)
    eventId = nil
  end
end


BestInSlot:RegisterMenuItem({
  descr = L["Overview"], 
  draw = MakeContainer, 
  width = 700,
  height = 600,
  close = onClose
})