local BestInSlot = LibStub("AceAddon-3.0"):GetAddon("BestInSlot")
local AceGUI = LibStub("AceGUI-3.0")
local L = LibStub("AceLocale-3.0"):GetLocale("BestInSlot")
local dropdownType, dropdownSelection, dropdownRequestFrom, dropdownDifficulty, dropdownFilter
local whisperTarget
local whisperTo
local requestButton
local responseContainer
local selectedType, selectedSelection, selectedChannel, selectedDifficulty
local cachedColor = "|cffb5b4ff"
local normalColor = "|cffb5ffb4"
local queuedCustomItems = {}
local history = {}
local filter = {}
local showObtainedItems = true
local classTokens = {
  VANQUISHER = {"ROGUE", "DEATHKNIGHT", "MAGE", "DRUID"},
  PROTECTOR = {"WARRIOR", "HUNTER", "SHAMAN", "MONK"},
  CONQUEROR = {"PALADIN", "PRIEST", "WARLOCK"},
}
local doRegister, setFilterOptions
local guildName

local function OnCustomItemInfoReply()
  for i=#queuedCustomItems,1, -1 do
    local widget = queuedCustomItems[i]
    local item = BestInSlot:GetItem(widget:GetUserData("itemid"))
    if item then
      local text = item.link
      if selectedType == "instance" then
        text = text.." - "..normalColor..item.source.."|r"
      end
      widget:SetUserData("itemlink", item.link)
      widget:SetText(text)
      table.remove(queuedCustomItems, i)
    end
  end
  if #queuedCustomItems == 0 then
    BestInSlot:UnregisterEvent("AddonMessage_requestCustomItemInfoReply", OnCustomItemInfoReply)
  end
end

local function addQueuedCustomItem(widget)
  if #queuedCustomItems == 0 then
    BestInSlot:RegisterEvent("AddonMessage_requestCustomItemInfoReply", OnCustomItemInfoReply)
  end
  table.insert(queuedCustomItems, widget)
end

local function helperColorCache(color)
  --Increase them all by 50% to make them lighter
  local r = color.r * 255
  local g = color.g * 255
  local b = color.b * 255
  r = BestInSlot:ToHexString(r + (255 - r) * 0.33)
  g = BestInSlot:ToHexString(g + (255 - g) * 0.33)
  b = BestInSlot:ToHexString(b + (255 - b) * 0.33)
  return  r..g..b
end

---
-- Helper function to format the labels properly
---
local function setText(label)
  local itemLink = label:GetUserData("itemlink")
  local chars = label:GetUserData("chars")
  local str = itemLink
  local first = true
  for k,v in pairs(chars) do
    if first then --comma or colon
      str = str..": "
      first = false
    else
      str = str..", "
    end
    if v then --color if cached or not
      str = str.."|cff"..helperColorCache(RAID_CLASS_COLORS[BestInSlot:GetPlayerClass(k)] or DEFAULT_CHATFRAME_COLOR)..k.."|r"
    else
      str = str..BestInSlot:GetPlayerString(k)
    end
  end
  label:SetText(str)
end

local function populateCache()
  if not responseContainer then return end --should this function be called when the responseContainer is allready closed
  local raidTier
  if selectedType == "perplayer" then
    raidTier = selectedSelection
  else
    raidTier = BestInSlot:GetRaidTiers(BestInSlot.TYPE_INSTANCE, selectedSelection)
  end
  local difficulty = selectedDifficulty
  if IsInGuild() then
    guildName = guildName or GetGuildInfo("player")
    local guildCache = BestInSlot.db.factionrealm[guildName]
    for charName, charData in pairs(guildCache) do
      if charData[raidTier] and (#charData[raidTier][difficulty] ~= 0 or (charData[raidTier][difficulty] and charData[raidTier][difficulty].spec)) then
        doRegister(charName, charData[raidTier][difficulty], true)
      end
    end
    setFilterOptions()
  end
end

local function resetPerPlayer(self)
  local characters = self:GetUserData("characters")
  if characters then wipe(characters) end
  self:ReleaseChildren()
end

local function registerPerPlayer(self, name, bis, cache)
  local characters = self:GetUserData("characters")
  local simpleGroup = characters[name]
  if simpleGroup then
    simpleGroup:ReleaseChildren()
  else 
    simpleGroup = AceGUI:Create("SimpleGroup")
    simpleGroup:SetFullWidth(true)
    characters[name] = simpleGroup
    self:AddChild(simpleGroup)
  end
  local heading = AceGUI:Create("Heading")
  heading:SetText(BestInSlot:GetPlayerString(name))
  heading:SetFullWidth(true)
  simpleGroup:AddChild(heading)
  for i, _ in pairs(bis) do
    local itemid, obtained = bis[i].item, bis[i].obtained
    if showObtainedItems or not bis[i].obtained then
      local label = BestInSlot:GetItemLinkLabel(itemid)
      label:SetFullWidth(true)
      local item = BestInSlot:GetItem(itemid)
      local text
      if item then
        text = item.link.." - "
      else
        text = "Custom item - "
        addQueuedCustomItem(label)
      end
      if cache then
        text = text..cachedColor
      else
        text = text..normalColor
      end
      if item then
        text = text..item.source
      end
      text = text .. "|r"
      label:SetText(text)  
      simpleGroup:AddChild(label)
    end
  end
  self:DoLayout()
end

local function populatePerPlayer()
  responseContainer:SetUserData("characters", {})
  responseContainer:SetLayout("List")
  responseContainer.Register = registerPerPlayer
  responseContainer.Reset = resetPerPlayer
end

local function fillBossGroups()
  local bosses = responseContainer:GetUserData("bosses")
  for i=1,bosses do
    local group = AceGUI:Create("SimpleGroup")
    group:SetFullWidth(true)
    local header = AceGUI:Create("Heading")
    header:SetFullWidth(true)
    header:SetText(BestInSlot:GetDescription(BestInSlot.TYPE_BOSS, responseContainer:GetUserData("selection"), i))
    group:AddChild(header)
    responseContainer:AddChild(group)
    responseContainer:SetUserData("boss"..i, group)
  end
end

local function resetRaidTier(self)
  responseContainer:ReleaseChildren()
  fillBossGroups()
end

local function registerRaidTier(self, name, bis, cache)
  cache = cache or false --make sure the value is false, nil won't be stored in the table, and we need it to be stored in the table for the name
  for i,_ in pairs(bis) do
    local itemid, obtained = bis[i].item, bis[i].obtained
    if showObtainedItems or not obtained then
      local item = BestInSlot:GetItem(itemid)
      if item and item.raidId == self:GetUserData("selection") then --item should be from this instance
        local bossGroup = self:GetUserData("boss"..item.bossid)
        if bossGroup then
          local label = bossGroup:GetUserData(itemid)
          if not label then
            local label = BestInSlot:GetItemLinkLabel(itemid)
            label:SetUserData("chars", {[name] = cache})
            label:SetFullWidth(true)
            setText(label)
            bossGroup:AddChild(label)
            bossGroup:SetUserData(itemid, label)
          else
            local chars = label:GetUserData("chars")
            chars[name] = cache
            setText(label)
          end
        end
      end
    end
  end
  self:DoLayout()
end

local function populateRaidTier()
  --TODO: Split the 'instance' type up in a left and right part 
  responseContainer:SetLayout("List")
  local bosses = BestInSlot:GetInstanceBosses(responseContainer:GetUserData("selection"))
  responseContainer:SetUserData("bosses", #bosses)
  responseContainer.Register = registerRaidTier
  responseContainer.Reset = resetRaidTier
  resetRaidTier()
end


local function populateResponseContainer()
  local responseType = responseContainer:GetUserData("type")
  if responseType == "perplayer" then
    populatePerPlayer()
  elseif responseType == "instance" then
    populateRaidTier()
  end
  ---
  -- Populate the responsecontainer with data from the cache
  ---
  if responseContainer:GetUserData("channel") == "GUILD" or not responseContainer:GetUserData("channel") then
    guildName = guildName or GetGuildInfo("player")
    local guildCache = BestInSlot.db.factionrealm[guildName]
    local type = responseContainer:GetUserData("type")
    local raidTier
    if type == "instance" then
      raidTier = BestInSlot:GetRaidTiers(BestInSlot.TYPE_INSTANCE, responseContainer:GetUserData("selection"))
    elseif type == "perplayer" then
      raidTier = responseContainer:GetUserData("selection")
    end
    if not raidTier then error("Something went wrong in fetching the raidTier") end
    local difficulty = responseContainer:GetUserData("difficulty")
    if not BestInSlot:IsRaidTierInitialized(raidTier) then
      BestInSlot:RegisterEvent("InitializeLootTable", populateCache)
      BestInSlot:InitializeLootTable(raidTier)
    else
      populateCache()
    end
  end
end

---Called when the actual request button is pressed, or to simulate it being pressed at startup
local function doRequest(cacheOnly)
  if type(cacheOnly) ~= "boolean" then
    cacheOnly = false
  end
  wipe(history)
  wipe(filter)
  if not responseContainer or not selectedType or not selectedSelection or not selectedDifficulty then return end
  responseContainer:Reset()
  responseContainer:ReleaseChildren()
  local msg = {selection = selectedSelection, difficulty = selectedDifficulty}
  responseContainer:SetUserData("type", selectedType)
  responseContainer:SetUserData("selection", selectedSelection)
  responseContainer:SetUserData("channel", selectedChannel)
  responseContainer:SetUserData("difficulty", selectedDifficulty)
  populateResponseContainer()
  local raidtier
  if selectedType == "perplayer" then
    raidtier = selectedSelection
  else
    raidtier = BestInSlot:GetRaidTiers(BestInSlot.TYPE_INSTANCE, selectedSelection)
  end
  local function dataRequest()
    if selectedType == "perplayer" then
      BestInSlot:SendAddonMessage("requestRaidTier", msg, selectedChannel, whisperTo)
    elseif selectedType == "instance" then
      BestInSlot:SendAddonMessage("requestRaidInstance", msg, selectedChannel, whisperTo)
    end
  end
  if not cacheOnly then
    if BestInSlot:IsRaidTierInitialized(raidtier) then
      dataRequest()
    else
      BestInSlot:RegisterEvent("InitializeLootTable", dataRequest)
      BestInSlot:InitializeLootTable(raidtier)
    end
  end
end

local function canRequest()
  if selectedType and selectedSelection and selectedDifficulty then
    if selectedChannel and (selectedChannel ~= "WHISPER" or whisperTo) then
      requestButton:SetDisabled(false)
    end
    doRequest(true)
  else
    requestButton:SetDisabled(true)
  end
end

local function setSelectionDropdown(_,_,type)
  if type ~= selectedType then
    responseContainer:Reset()
    responseContainer:ReleaseChildren()
    if selectedType == "perplayer" then
      selectedSelection = BestInSlot:GetLatest(BestInSlot.TYPE_INSTANCE, BestInSlot.TYPE_RAIDTIER, selectedSelection)
    elseif selectedType == "instance" then
      selectedSelection = BestInSlot:GetRaidTiers(BestInSlot.TYPE_INSTANCE, selectedSelection)
    end
  end
  selectedType = type
  local newSelection = {}
  if type == "instance" then
    local raids = BestInSlot:GetInstances()
    for i=1,#raids do
      newSelection[raids[i]] = BestInSlot:GetDescription(BestInSlot.TYPE_INSTANCE, raids[i])
    end
    dropdownSelection:SetList(newSelection)
    dropdownSelection:SetDisabled(false)
    dropdownSelection:SetText("Select an instance")
    dropdownSelection:SetLabel("Instance")
  elseif type == "perplayer" then
    local raidTiers = BestInSlot:GetRaidTiers()
    for i=1,#raidTiers do
      newSelection[raidTiers[i]] = BestInSlot:GetDescription(BestInSlot.TYPE_RAIDTIER, raidTiers[i])
    end
    dropdownSelection:SetList(newSelection)
    dropdownSelection:SetDisabled(false)
    dropdownSelection:SetText("Select a raid tier")
    dropdownSelection:SetLabel("Raid Tier")
  end
  if selectedSelection then
    dropdownSelection:SetValue(selectedSelection)
  else
    dropdownSelection:SetValue(nil)
    dropdownDifficulty:SetDisabled(true)
    dropdownDifficulty:SetList({})
    dropdownDifficulty:SetValue(nil)
  end
  canRequest()
end

local function setDifficultyDropdown(_,_,value)
  selectedSelection = value
  local type
  if selectedType == "perplayer" then
    type = BestInSlot.TYPE_RAIDTIER
  else  
    type = BestInSlot.TYPE_INSTANCE
  end
  local list = BestInSlot:GetDifficulties(type, value)
  if selectedDifficulty and selectedDifficulty > #list then
    selectedDifficulty = 1
  end
  dropdownDifficulty:SetList(BestInSlot:GetDifficulties(type, value))
  dropdownDifficulty:SetDisabled(false)
  dropdownDifficulty:SetValue(selectedDifficulty)
  canRequest()
end

local function showTooltip(widget)
  local itemid = widget:GetUserData("itemid")
  BestInSlot:GetItemTooltip(itemid, BestInSlot.frame)
end

local function hideTooltip()
  BestInSlot:HideItemTooltip()
end

local function releaseResponseContainer(self)
  self:ReleaseChildren()
end

local function registerResponseContainer() end

local function drawMenu(container)
  container:SetLayout("Flow")
  dropdownType = AceGUI:Create("Dropdown")
  dropdownType:SetLabel(L["Request type"])
  dropdownType:SetList({
    perplayer = L["BestInSlot per Player"],
    instance = L["Instance Overview"],
  })
  dropdownType:SetValue("instance")
  selectedType = selectedType or "instance"
  dropdownType:SetRelativeWidth(0.33)
  dropdownType:SetText(L["Select a request type"])
  dropdownType:SetCallback("OnValueChanged", setSelectionDropdown)

  container:AddChild(dropdownType)
  
  dropdownSelection = AceGUI:Create("Dropdown")
  dropdownSelection:SetLabel(L["Selection"])
  dropdownSelection:SetRelativeWidth(0.33)
  dropdownSelection:SetText(L["Select a type first!"])
  dropdownSelection:SetList({nothing = " "})
  dropdownSelection:SetDisabled(true)
  dropdownSelection:SetCallback("OnValueChanged", setDifficultyDropdown)
  
  container:AddChild(dropdownSelection)
  
  dropdownDifficulty = AceGUI:Create("Dropdown")
  dropdownDifficulty:SetLabel(L["Difficulty"])
  dropdownDifficulty:SetRelativeWidth(0.33)
  dropdownDifficulty:SetDisabled(true)
  dropdownDifficulty:SetCallback("OnValueChanged", function(_,_,value) 
    selectedDifficulty = value
    canRequest()  
  end)
  
  container:AddChild(dropdownDifficulty)
  
  dropdownRequestFrom = AceGUI:Create("Dropdown")
  dropdownRequestFrom:SetLabel(L["Request from"])
  dropdownRequestFrom:SetRelativeWidth(0.221)
  dropdownRequestFrom:SetList({
    RAID = ("%s/%s"):format(_G["PARTY"], _G["RAID"]),
    GUILD = _G["GUILD"],
    WHISPER = _G["WHISPER"],    
  })
  if not IsInGuild() then
    dropdownRequestFrom:SetItemDisabled("GUILD", true)
  end
  if not IsInGroup() then
    dropdownRequestFrom:SetItemDisabled("RAID", true)
  end
  dropdownRequestFrom:SetText(L["Select a channel"])
  dropdownRequestFrom:SetCallback("OnValueChanged", function(_,_,value)
    selectedChannel = value
    if value == "WHISPER" then
      whisperTarget:SetDisabled(false)
    else
      whisperTarget:SetDisabled(true)
    end
    canRequest()
  end)
  
  container:AddChild(dropdownRequestFrom)
  
  whisperTarget = AceGUI:Create("EditBox")
  whisperTarget:SetLabel(L["Whisper target"])
  whisperTarget:SetRelativeWidth(0.221)
  whisperTarget:SetDisabled(true)
  whisperTarget:SetCallback("OnEnterPressed", function(_,_,value)
    whisperTo = value
    canRequest()  
  end)
  
  container:AddChild(whisperTarget)
  
  requestButton = AceGUI:Create("Button")
  requestButton:SetText(L["Request"])
  requestButton:SetRelativeWidth(0.221)
  requestButton:SetDisabled(true)
  requestButton:SetCallback("OnClick", doRequest)
    
  container:AddChild(requestButton)
  
  if selectedType then
    dropdownType:SetValue(selectedType)
    setSelectionDropdown(nil, nil, selectedType)
    setDifficultyDropdown(nil, nil, selectedSelection)
    dropdownRequestFrom:SetValue(selectedChannel)
    if selectedChannel == "WHISPER" then
      whisperTarget:SetDisabled(false)
    end
    whisperTarget:SetText(whisperTo)
  end
  
  dropdownFilter = AceGUI:Create("Dropdown")
  dropdownFilter:SetDisabled(true)
  dropdownFilter.dropdown.obj:SetDisabled(true)
  dropdownFilter:SetRelativeWidth(0.33)
  dropdownFilter:SetList({})
  dropdownFilter:SetLabel(FILTER)
  
  container:AddChild(dropdownFilter)
  
  responseContainer = AceGUI:Create("ScrollFrame")
  responseContainer:SetLayout("List")
  responseContainer:SetFullWidth(true)
  responseContainer:SetFullHeight(true)
  responseContainer:SetPoint("BOTTOMRIGHT", container.frame, "BOTTOMRIGHT")
  responseContainer.Reset = releaseResponseContainer
  responseContainer.Register = registerResponseContainer
    
  container:AddChild(responseContainer)
  canRequest()
end

local function close()
  dropdownType = nil
  dropdownSelection = nil
  dropdownRequestFrom = nil
  dropdownDifficulty = nil
  dropdownFilter = nil
  whisperTarget = nil
  requestButton = nil
  responseContainer = nil
  wipe(queuedCustomItems)
  wipe(history)
end

local function sendRaidTierReply(raidTier, difficulty, source, version)
  local bisList
  if version >= 1 and version <= 96 then -- The >= 1 is ment for development versions
    local specId = GetSpecializationInfo(BestInSlot:GetSpecialization())
    bisList = BestInSlot:GetBestInSlotItems(raidTier, difficulty, specId)
  else
    bisList = BestInSlot:GetBestInSlotItems(raidTier, difficulty) 
  end
  BestInSlot:SendAddonMessage("requestRaidTierReply", bisList, "WHISPER", source)
end

local function requestRaidTier(data, channel, source, version)
  if not data.selection or not data.difficulty then BestInSlot.console:AddError("Invalid data received for requestRaidTier") return end
  local raidTier = data.selection
  local difficulty = data.difficulty
  sendRaidTierReply(raidTier, difficulty, source, version)
end

local function requestRaidInstance(data, channel, source, version)
  if not data.selection or not data.difficulty then BestInSlot.console:AddError("Invalid data received for requestRaidInstance") return end
  local instance = data.selection
  local difficulty = data.difficulty
  local raidTier = BestInSlot:GetRaidTiers(BestInSlot.TYPE_INSTANCE, instance)
  sendRaidTierReply(raidTier, difficulty, source, version)
end

local function emptySetValue() end --Not used for dropdown-item-menu, therefore a filler empty function

local function applyNewFilter()
  responseContainer:Reset()
  for char,data in pairs(history) do
    if filter[char].show then
      if data.spec then
        for specId, specData in pairs(data) do
          if specId ~= "spec" and filter[char][specId] then
            responseContainer:Register(char, data[specId], true)
          end
        end
      else
        responseContainer:Register(char, data, true)
      end
    end
  end
end

---Verifies the token filter menu with what is actually in the filter table
local function checkTokenDisable()
  for tokenType, tokenClasses in pairs(classTokens) do
    local tokenSelector = dropdownFilter:GetUserData(tokenType)
    if tokenSelector then
      local tokenEnabled = false
      for i=1,#tokenClasses do
        local classSelector = dropdownFilter:GetUserData(tokenClasses[i])
        if classSelector and classSelector:GetValue() then
          tokenEnabled = true
          break
        end
      end
      tokenSelector:SetValue(tokenEnabled)
    end
  end
end

---Verifies the class filter menu with what is actually in the filter table
local function checkClassDisable(recursive)
  recursive = recursive or true
  local hasChanged = false
  for class in pairs(RAID_CLASS_COLORS) do -- we use the RAID_CLASS_COLORS global since that contains all the classes
    local classWidget = dropdownFilter:GetUserData(class)
    if classWidget then
      local shouldShowClass = false
      for char in pairs(filter) do
        if BestInSlot:GetPlayerClass(char) == class then
          if filter[char].show then
            shouldShowClass = true
            break
          end
        end
      end
      if classWidget:GetValue() ~= shouldShowClass then
        hasChanged = true
        classWidget:SetValue(shouldShowClass)
      end
    end
  end
  if hasChanged and recursive then
    checkTokenDisable()
  end
end

---Verifies the guild ranks filter menu with what is actually in the filter table
local function checkGuildRankDisable(recursive)
  recursive = recursive or true
  for i=1,GuildControlGetNumRanks() do
    local guildWidget = dropdownFilter:GetUserData("GuildRank"..i)
    if guildWidget then 
      guildWidget:SetValue(false)
      guildWidget:SetDisabled(true) 
    end --Disable it before looping through the roster
  end
  for i=1,GetNumGuildMembers() do
    local fullName, _, rankIndex = GetGuildRosterInfo(i)
    rankIndex = rankIndex + 1
    local playerName = (fullName):match("(%D+)-(%D+)")
    if filter[playerName] then --Player is known in our filter
      local guildWidget = dropdownFilter:GetUserData("GuildRank"..rankIndex) --get the widget for the player's guild rank
      if guildWidget  then
        if filter[playerName].show then
          guildWidget:SetValue(true) -- Tick the box if we have a player shown off this guild rank
        end
        guildWidget:SetDisabled(false) --Enable the tickbox for use
      end
    end
  end
end

---Verifies the character filter menu with what is actually in the filter table
local function checkPlayerDisable(recursive)
  recursive = recursive or true
  local hasChanged = false
  for char, charFilter in pairs(filter) do
    local charGroup = dropdownFilter:GetUserData(char)
    local widgetGroup = charGroup:GetUserDataTable()
    for widgetDescr,widget in pairs(widgetGroup) do
      if widgetDescr == "visible" then --if it's the show widget
        if widget:GetValue() ~= (charFilter.show or false) then
          hasChanged = true
        end
        widget:SetValue(charFilter.show or false)          
      elseif widget then --if it's a spec widget
        widget:SetValue(charFilter[widgetDescr] or false)
      end
    end
  end
  if hasChanged and recursive then
    checkClassDisable()
  end
end

local function verifyCheckBoxes()
  checkPlayerDisable(false)
  checkGuildRankDisable(false)
  checkClassDisable(false)
  checkTokenDisable(false)
end

--- Simply (un)ticks all the options in the filter
-- @param #boolean value Wether to tick, or untick everything
-- @param #boolean ignoreApply Optional, if true then doesn't apply the filter afterwards. 
local function setAllFilter(value, ignoreApply)
  local userData = dropdownFilter:GetUserDataTable()
  for char, charFilter in pairs(filter) do
    charFilter.show = value
  end
  verifyCheckBoxes()
  if not ignoreApply then
    applyNewFilter()
  end
end

--- Untick the character, and checks if other tickboxes should be unticked as a result
-- @param #string char Character name
-- @param #boolean value Was the character enabled (true) or disabled (false)
-- @param #boolean ignoreApply Optional, if true then doesn't apply the filter afterwards.
local function setCharFilter(char, value, ignoreApply)
  local userData = dropdownFilter:GetUserDataTable()
  for filterType in pairs(filter[char]) do
    if filterType == "show" then
      filter[char][filterType] = value
    elseif (not value) or (history[char] and history[char].spec == filterType)  then --Set all specs to false when unticked, set active spec to true when ticked
      filter[char][filterType] = value
    end
  end
  filter[char].show = value
  local class = BestInSlot:GetPlayerClass(char)
  local playerClassTickerEnabled = false
  for char, charData in pairs(filter) do
    if BestInSlot:GetPlayerClass(char) == class and charData.show then --There is still a char with the same class enabled
      playerClassTickerEnabled = true
      break
    end
  end
  if userData[class] then userData[class]:SetValue(playerClassTickerEnabled) end
  checkClassDisable()
  checkGuildRankDisable()
  if not ignoreApply then
    applyNewFilter()
  end
end

local function setSpecFilter(char, spec, value, ignoreApply)
  local shouldRefresh = not value --when ticking a box, it defualts to not show
  filter[char][spec] = value
  if value then
    filter[char].show = true
    shouldRefresh = true
  else
    for k,v in pairs(filter[char]) do
      if k ~= "show" and v then -- Check if a spec is still selected, if none
        shouldRefresh = false
        break
      end
    end
    if shouldRefresh then
      filter[char].show = false
    end
  end
  if shouldRefresh then
    verifyCheckBoxes()
  end 
  if not ignoreApply then
    applyNewFilter()
  end
end

--- Untick the class, and check if other tickboxes should be unticked as a result
-- @param #string class Unlocalized class string
-- @param #boolean value Was the class enabled (true) or disabled (false)
-- @param #boolean ignoreApply Optional, if true then doesn't apply the filter afterwards.
local function setClassFilter(class, value, ignoreApply)
  local userData = dropdownFilter:GetUserDataTable()
  for char, data in pairs(filter) do
    if BestInSlot:GetPlayerClass(char) == class then
      data.show = value
      if userData[char] then
        local visible = userData[char]:GetUserData("visible")
        visible:SetValue(value)
      end
    end
  end
  if not ignoreApply then
    applyNewFilter()
  end
  checkTokenDisable()
  checkGuildRankDisable()
end

--- On clicking a guild filter it'll update the filter according to the guild rank that has been unticked
-- @param #number guildRank the number of the guildrank that is to be unticked
-- @param #boolean value True if enabled, otherwise false
-- @param #boolean ignoreApply Optional, if true then doesn't apply the filter afterwards.
local function setGuildFilter(guildRank, value, ignoreApply)
  for i=1,GetNumGuildMembers() do
    local fullName, _, rankIndex = GetGuildRosterInfo(i)
    rankIndex = rankIndex + 1
    if rankIndex == guildRank then
      local playerName = (fullName):match("(%D+)-(%D+)")
      if filter[playerName] then
        filter[playerName].show = value
      end
    end
  end
  if not ignoreApply then
    applyNewFilter()
  end
  checkPlayerDisable()
end


local function onClassSelect(widget, _, value)
  local class = widget:GetUserData("class")
  setClassFilter(class, value)
end

local function onTokenSelect(widget, _, value)
  local token = widget:GetUserData("token")
  for i=1,#classTokens[token] do
    local classSelect = dropdownFilter:GetUserData(classTokens[token][i])
    if classSelect then
      classSelect:SetValue(value)
      setClassFilter(classTokens[token][i], value, true) --simulate that we're clicking all 3 or 4 class tokens buttons
    end
  end
  applyNewFilter()
end

local function onClickSelectAll(widget)
  local selected = widget:GetUserData("select") or false
  setAllFilter(selected)
end

local function onCharacterView(widget, _, value)
  local char = widget:GetUserData("char")
  setCharFilter(char, value)
end

local function onSpecChanged(widget, _, value)
  local char = widget:GetUserData("char")
  local spec = widget:GetUserData("spec")
  setSpecFilter(char, spec, value)
end

local function onGuildRankSelect(widget, _, value)
  local guildRankId = widget:GetUserData("rank")
  setGuildFilter(guildRankId, value)
end

local function getDropdownSeperator()
  local seperator = AceGUI:Create("Dropdown-Item-Separator")
  seperator.SetValue = emptySetValue
  return seperator
end

local function onShowObtainedValue(widget, _, value)
  showObtainedItems = value
  applyNewFilter()
end

local function makeStandardOptions(submenu)
  local hasItems = false
  for _ in pairs(filter) do
    hasItems = true
    break
  end
  local isGuildRequest = responseContainer:GetUserData("channel") == "GUILD" or not responseContainer:GetUserData("channel")
  if not hasItems or not isGuildRequest then
    if not isGuildRequest then --Don't show anything for party/whisper requests
      return
    end
    local header = AceGUI:Create("Dropdown-Item-Header")
    header:SetText(L["Not enough data to filter"])
    header.SetValue = emptySetValue
    submenu:AddItem(header)
    return
  end
  -- Add Select All button
  local selectAll = AceGUI:Create("Dropdown-Item-Execute")
  selectAll:SetText(L["Select all"])
  selectAll:SetCallback("OnClick", onClickSelectAll)
  selectAll.SetValue = emptySetValue
  selectAll:SetUserData("select", true)
  submenu:AddItem(selectAll)
  
  --Add Deselect All button
  local deselectAll = AceGUI:Create("Dropdown-Item-Execute")
  deselectAll:SetText(L["Deselect all"])
  deselectAll:SetCallback("OnClick", onClickSelectAll)
  deselectAll.SetValue = emptySetValue
  submenu:AddItem(deselectAll)
  
  local showObtained = AceGUI:Create("Dropdown-Item-Toggle")
  showObtained:SetText(L["Show obtained items"])
  showObtained:SetValue(showObtainedItems or true)
  showObtained:SetCallback("OnValueChanged", onShowObtainedValue)
  submenu:AddItem(showObtained)
  
  submenu:AddItem(getDropdownSeperator())
  
  --Add the class Menu
  local classMenu = AceGUI:Create("Dropdown-Item-Menu")
  classMenu:SetText(L["Class filter"])
  classMenu.SetValue = emptySetValue
  local classSubMenu = AceGUI:Create("Dropdown-Pullout")
  classSubMenu.SetValue = emptySetValue
  local classesDone = {}
  for charName, charData in pairs(filter) do
    local class, localizedClass = BestInSlot:GetPlayerClass(charName)
    if not classesDone[class] then
      local classToggle = AceGUI:Create("Dropdown-Item-Toggle")
      classToggle:SetUserData("class", class)
      classToggle:SetText(BestInSlot:GetClassString(class))
      classToggle:SetCallback("OnValueChanged", onClassSelect)
      dropdownFilter:SetUserData(class, classToggle)
      classesDone[class] = classToggle
      classSubMenu:AddItem(classToggle)
    end
  end
  classMenu:SetMenu(classSubMenu)
  submenu:AddItem(classMenu)
  
  --Add the token menu
  local tokenMenu = AceGUI:Create("Dropdown-Item-Menu")
  tokenMenu:SetText(L["Token filter"])
  tokenMenu.SetValue = emptySetValue
  local tokenSubMenu = AceGUI:Create("Dropdown-Pullout")
  tokenSubMenu.SetValue = emptySetValue
  tokenSubMenu:SetWidth(300)
  local tokensDone = {}
  for tokenType, classes in pairs(classTokens) do
    for i=1, #classes do
      if classesDone[classes[i]] then
        local tokenToggle = tokensDone[tokenType] or AceGUI:Create("Dropdown-Item-Toggle")
        if not tokensDone[tokenType] then
          tokenToggle:SetUserData("token", tokenType)
          local text = L[tokenType].." ("
          for j=1,#classes do
            if j ~= 1 then
              text = text.." / "
            end
            text = text..BestInSlot:GetClassString(classes[j])
          end
          text = text..")"
          tokenToggle:SetText(text)
          tokenToggle:SetCallback("OnValueChanged", onTokenSelect)
          dropdownFilter:SetUserData(tokenType, tokenToggle)
          tokensDone[tokenType] = tokenToggle
          tokenSubMenu:AddItem(tokenToggle)
          break
        end
      end
    end
  end
  tokenMenu:SetMenu(tokenSubMenu)
  submenu:AddItem(tokenMenu)
  
  --Adds the guild filter
  if responseContainer:GetUserData("channel") == "GUILD" or not responseContainer:GetUserData("channel") then -- if the guild window is shown
    local guildMenu = AceGUI:Create("Dropdown-Item-Menu")
    guildMenu:SetText(GUILDCONTROL_GUILDRANKS)
    guildMenu.SetValue = emptySetValue
    local guildSubMenu = AceGUI:Create("Dropdown-Pullout")
    guildSubMenu.SetValue = emptySetValue
    for i=1, GuildControlGetNumRanks() do
      local guildRank = AceGUI:Create("Dropdown-Item-Toggle")
      guildRank:SetText(GuildControlGetRankName(i))
      guildRank:SetUserData("rank", i)
      guildRank:SetCallback("OnValueChanged", onGuildRankSelect)
      guildSubMenu:AddItem(guildRank)
      dropdownFilter:SetUserData("GuildRank"..i, guildRank)
    end
    guildMenu:SetMenu(guildSubMenu)
    submenu:AddItem(guildMenu)
  end
  submenu:AddItem(getDropdownSeperator())
end



local function makeCharSpecificOptions(submenu, char)
  local header = AceGUI:Create("Dropdown-Item-Header")
  header:SetText(BestInSlot:GetPlayerString(char))
  submenu:AddItem(header)
  local visibleItem = AceGUI:Create("Dropdown-Item-Toggle")
  visibleItem:SetText(SHOW)
  visibleItem:SetUserData("char", char)
  visibleItem:SetCallback("OnValueChanged", onCharacterView)
  submenu:SetUserData("visible", visibleItem)
  submenu:AddItem(visibleItem)
  if history[char] and history[char].spec then
    submenu:AddItem(getDropdownSeperator())
    
    for specId in pairs(history[char]) do
      if type(tonumber(specId)) == "number" then
        local _, specName = GetSpecializationInfoByID(specId)
        local specOptions = AceGUI:Create("Dropdown-Item-Toggle")
        specOptions:SetText(specName)
        specOptions:SetUserData("spec", tonumber(specId))
        specOptions:SetUserData("char", char)
        specOptions:SetCallback("OnValueChanged", onSpecChanged)
        submenu:AddItem(specOptions)
      end
    end
  end
end

local function validateFilters()
  for k,v in pairs(filter) do
    if not history[k] then
      filter[k] = nil
    end
  end
end

setFilterOptions = function()
  dropdownFilter:SetList({})
  validateFilters()
  wipe(dropdownFilter:GetUserDataTable())
  makeStandardOptions(dropdownFilter.pullout)
  local disabled = true
  local orderedList = {}
  for char in pairs(filter) do
    if not history[char] then
      filter[char] = nil
    else
      table.insert(orderedList, char)
    end
  end
  table.sort(orderedList)
  for i,char in pairs(orderedList) do
    disabled = false
    local submenu = AceGUI:Create("Dropdown-Pullout")
    makeCharSpecificOptions(submenu, char)
    
    local dropdownGroup = AceGUI:Create("Dropdown-Item-Menu")
    dropdownGroup:SetText(BestInSlot:GetPlayerString(char))
    dropdownGroup:SetMenu(submenu)
    dropdownGroup.SetValue = emptySetValue
    
    dropdownFilter.pullout:AddItem(dropdownGroup)
    dropdownFilter:SetUserData(char, submenu)
  end
  dropdownFilter:SetDisabled(disabled)
  verifyCheckBoxes()
end

local function formatResponse(source, bisList)
  guildName = guildName or GetGuildInfo("player")
  if type(bisList[1]) == "table" or bisList.spec then
    if bisList.spec then
      local activeSpec = BestInSlot.db.factionrealm[guildName][source].activeSpec or bisList.spec
      history[source] = bisList
      filter[source] = filter[source] or {show = true, [activeSpec] = true}
      return bisList[bisList.spec]
    elseif bisList[1] and bisList[1].item then
      history[source] = bisList
      filter[source] = filter[source] or {show = true}
      return bisList
    end
  elseif type(bisList[1]) == "number" then
    local newTable = {}
    for i=1,#bisList do
      newTable[i] = {item = bisList[i], obtained = false}
    end
    filter[source] = filter[source] or {show = true}
    history[source] = newTable
    return newTable
  end
  if #bisList ~= 0 then
    BestInSlot.console:AddError("Invalid data for registering responsecontainer", source, bisList)
  end
  return {}
end

doRegister = function(source, bisList, cache)
  bisList = formatResponse(source, bisList)
  for i=1,#bisList do
    if not cache and bisList[i].isCustom and not BestInSlot:IsItemCustom(bisList[i].item) then
      BestInSlot:SendAddonMessage("requestCustomItemInfo", bisList[i].item, "WHISPER", source)
    end
  end
  responseContainer:Register(source, bisList, cache)
  if not cache then
    setFilterOptions() --only do this when it's not from the cache, cause if it is it'll probably spam like shit!
  end
end

---
-- Itemlist format: 
-- {
--  [specId1] = {{item = item1, obtained = false}, {item = item2, obtained = true}}
--  [specId2] = {{item = item1, obtained = false}, {item = item2, obtained = true}}
--  spec = CurrentSpec (used for the other client to define what to show)
-- }
---
local function requestRaidTierReply(itemlist, channel, source)
  if not responseContainer then return end
  guildName = guildName or GetGuildInfo("player")
  local raidTier
  local requestType = responseContainer:GetUserData("type")
  if requestType == "instance" then
    raidTier = BestInSlot:GetRaidTiers(BestInSlot.TYPE_INSTANCE, responseContainer:GetUserData("selection"))
  elseif requestType == "perplayer" then
    raidTier = responseContainer:GetUserData("selection")
  end
  if responseContainer:GetUserData("channel") == "GUILD" then
    BestInSlot.db.factionrealm[guildName][source][raidTier][responseContainer:GetUserData("difficulty")] = itemlist
    BestInSlot.db.factionrealm[guildName][source].activeSpec = itemlist.spec
  elseif channel == "WHISPER" then
    local list = BestInSlot:GetCacheData()
    if #list ~= 0 and list[source] and list[source].guild == guildName then
      BestInSlot.db.factionrealm[guildName][source][raidTier][responseContainer:GetUserData("difficulty")] = itemlist
      BestInSlot.db.factionrealm[guildName][source].activeSpec = itemlist.spec
    end
  end
  doRegister(source, itemlist)
  setFilterOptions()
end

local function requestCustomItemInfo(itemid, channel, source)
  local response = BestInSlot.db.global.customItems[itemid]
  if not response or not response.raidTier then
    response = nil
  end
  response = {id = itemid, data = response}
  BestInSlot:SendAddonMessage("requestCustomItemInfoReply", response, "WHISPER", source)
end

local function requestCustomItemInfoReply(data, channel, source)
  BestInSlot:AddCustomItem(data.id, data.data.type, data.data.raidTier, data.data.difficulty, data.data.dungeon, data.data.boss)
end

BestInSlot:RegisterCommFunction("requestRaidTier", requestRaidTier)
BestInSlot:RegisterCommFunction("requestRaidInstance", requestRaidInstance)
BestInSlot:RegisterCommFunction("requestRaidTierReply", requestRaidTierReply)
BestInSlot:RegisterCommFunction("requestCustomItemInfo", requestCustomItemInfo)
BestInSlot:RegisterCommFunction("requestCustomItemInfoReply", requestCustomItemInfoReply)
BestInSlot:RegisterEvent("PLAYER_GUILD_UPDATE", function()
  if IsInGuild() then
    guildName = GetGuildInfo("player")
  else
    guildName = nil
  end
end)

BestInSlot:RegisterMenuItem({
  descr = L["Request BiS"],
  draw = drawMenu,
  width = 700,
  height = 600,
  close = close,
})